/**
* @example AMDemodulationExample.cpp
*
* Demonstrates the setup of the AMDemodulator processor and its usage.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>
#include <math.h>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"

#undef _TempFile
#undef _DefaultFile
#ifdef _WIN32
#define _TempFile		"C:\\tmp\\libtrf-audioout"
#define _DefaultFile	"C:\\tmp\\AMOut.wav"
#else
#define _TempFile		"~/libtrf-audioout"
#define _DefaultFile	"~/tmp/AMOut.wav"
#endif

#define kAudioRateHz	(44.1e3f)


// Processor example code
// This example code opens a ThinkRF device, creates a Spectrum Stream and
// an AM Demodulator processor, streaming the output to an audio file.
//-----------------------------------------------------------------------------
bool AMDemodulationExample(int argc, const char *argv[]) {
	cout << "AMDemodulationExample" << endl;

	// Interpret target frequency range
	_float32 fCentreHz(1030.0e3f), fIFBWHz(100.0e3f);
	if (argc >= 2) {
		fCentreHz = strtof(argv[0], nullptr);
		fIFBWHz = strtof(argv[1], nullptr);
		cout << "Interpreted FCentre: " << fCentreHz << ", IFBW: " << fIFBWHz << endl;
	}

	// Interpret target filename
	const char *pOutputFile(_DefaultFile);
	if (argc >= 3) {
		pOutputFile = argv[2];
	}

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Grab AM station (Ottawa - 1310kHz)
	trfHandle hIQStream(trfInvalidHandle);
	eStatus = trfCreateIQStream(&hIQStream, fCentreHz, fIFBWHz, -30.0f);
	if (eStatus != trfOk) {
		cout << "trfCreateIQStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	//! [trfGetProcessorTypes]
	//!
	// Get the list of available processor types
	char *pTypes(nullptr);
	eStatus = trfGetProcessorTypes(&pTypes);
	cout << "trfGetProcessorTypes returns: " << endl << pTypes << endl;
	trfDisposeString(&pTypes);
	//! [trfGetProcessorTypes]

	//! [trfCreateProcessor]
	//!
	// Instantiate an AM demodulator
	trfHandle hProcessor(trfInvalidHandle);
	eStatus = trfCreateProcessor(&hProcessor, "AMDemodulator");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfCreateProcessor]

	//! [trfGetParameterInfo]
	//!
	// Display demodulator parameter set
	char *pParameterInfo(nullptr), *pParameters(nullptr);
	eStatus = trfGetParameterInfo(hProcessor, &pParameterInfo);
	cout << "Parameter info : " << endl << pParameterInfo << endl;
	//! [trfGetParameterInfo]

	//! [trfSetParameters]
	//!
	// Set AM demodulator parameters (obtain using trfGetParameterInfo)
	trfAddNumericParameter(&pParameters, TRFOutputSampleRate, kAudioRateHz);//  Program output sample rate - 44.1kHz
	trfAddTextParameter(&pParameters, TRFAMSubtype, "DSB");// Program AM demodulator type
	eStatus = trfSetParameters(hProcessor, &pParameters);
	if (eStatus != trfOk) {
		cout << "trfSetParameters failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	trfDisposeString(&pParameters);
	//! [trfSetParameters]

	//! [trfGetProcessorOutputStream]
	//!
	// Obtain baseband output stream from processor (AM Demodulator)
	trfHandle hAudioOut(trfInvalidHandle);
	eStatus = trfGetProcessorOutputStream(hProcessor, &hAudioOut);
	if (eStatus != trfOk) {
		cout << "trfGetProcessorOutputStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfGetProcessorOutputStream]

	// Attach the elements into a processing chain
	eStatus = trfAttachStreamToDevice(hDevice, hIQStream);
	if (eStatus != trfOk) {
		cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	//! [trfAttachProcessorToStream]
	//!
	// Attach hProcessor to its source IQ stream
	eStatus = trfAttachProcessorToStream(hProcessor, hIQStream);
	if (eStatus != trfOk) {
		cout << "trfAttachProcessorToStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfAttachProcessorToStream]

	// Attach the output stream to a file
	remove(_TempFile ".json");
	remove(_TempFile ".data");
	eStatus = trfSetStreamOutputFile(hAudioOut, _TempFile);
	if (eStatus != trfOk) {
		cout << "trfSetStreamOutputFile failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Loop for 5 seconds worth
	//! [trfGetNextBaseband]
	//!
	_uint32 uSamplesRemaining((_uint32)roundf(1.0f * kAudioRateHz));
	while(uSamplesRemaining > 0) {
		trfBasebandFrame *pFrame(nullptr);
		while (true) {
			eStatus = trfGetNextBaseband(hAudioOut, &pFrame, 1000);
			if (eStatus == trfOk) {
				break;
			}
		}
		if (pFrame->pJSONInfo != nullptr) {
			cout << "Annotation JSON: " << pFrame->pJSONInfo << endl;
		}
		uSamplesRemaining = (pFrame->uSamples > uSamplesRemaining) ? 0: (uSamplesRemaining - pFrame->uSamples);
		cout << "Samples remaining:" << uSamplesRemaining << endl;
		trfFreeBaseband(&pFrame);
	}
	//! [trfGetNextBaseband]

	eStatus = trfSetStreamOutputFile(hAudioOut, nullptr);

	trfGetParameters(hProcessor, &pParameters);
	cout << "Parameters : " << endl << pParameters << endl;
	trfDisposeString(&pParameters);

	//! [trfDetachProcessor]
	//!
	// Detach processor from its source stream
	eStatus = trfDetachProcessor(hProcessor);
	if (eStatus != trfOk) {
		cout << "Whups! " << eStatus << endl;
	}
	//! [trfDetachProcessor]

	trfDetachStream(hIQStream);

	trfClose(&hAudioOut);
	trfClose(&hProcessor);
	trfClose(&hIQStream);
	trfClose(&hDevice);

	//! [trfConvertBasebandFileToWavFile]
	//!
	const char *pBasebandFilePath(_TempFile);
	eStatus = trfConvertBasebandFileToWavFile(pBasebandFilePath, pOutputFile);
	if (eStatus != trfOk) {
		cout << "trfConvertBasebandFileToWavFile failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfConvertBasebandFileToWavFile]

	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	AMDemodulationExample(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
