/**
* @example CaptureToFileExample.cpp
*
* An example that demonstrates capturing IQ and spectrum data to file.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/


#include <iostream>
#include <string>
#include <cstring>
#include <chrono>
#include <thread>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#endif

// Include libtrf header file
#include "libtrf.h"

#ifndef _MAIN_
static std::string _goodFilename(std::string sSourceName) {
#ifdef _WIN32
	return sSourceName;
#else
	const char *pHomeDir(getenv("HOME"));
	if (nullptr == pHomeDir) {
		pHomeDir = getpwuid(getuid())->pw_dir;
	}
	std::string sHomeDir(pHomeDir);
	if (sSourceName[0] = '~') {
		sSourceName = sHomeDir + sSourceName.substr(1);
	}
	return sSourceName;
#endif
}

static void _removeOutputFiles(const char *pRootFilename) {
	char pFullFilename[256];
	sprintf(pFullFilename, "%s.json", pRootFilename);
	remove(pFullFilename);
	sprintf(pFullFilename, "%s.data", pRootFilename);
	remove(pFullFilename);
}
#endif


// Set of tests to run
const static struct {
	const char *pFilename;
	bool bSpectrumNotIQ;
	union {
		struct {
			_float32 fMinHz;
			_float32 fMaxHz;
			_float32 fRBWHz;
		}spectrum;
		struct {
			_float32 fCentreHz;
			_float32 fIFBWHz;
		}iq;
	}data;
	_uint32 uFrames;
}spTests[] = {
#ifdef _WIN32
	{"C:\\tmp\\spectrum-100MHz-1GHz-10kHz", true, {100.0e6f, 1.0e9f, 10.0e3f}, 100},
	{"C:\\tmp\\spectrum-1GHz-8GHz-100kHz", true, {1.0e9f, 8.0e9f, 100.0e3f}, 100},
	{"C:\\tmp\\iq-100MHz-10MHz", false, {100.0e6f, 10.0e6f}, 1000},
	{"C:\\tmp\\iq-1GHz-40MHz", false, {1.0e9f, 40.0e6f}, 100},
#else
	{"~/spectrum-100MHz-1GHz-10kHz", true, {100.0e6f, 1.0e9f, 10.0e3f}, 100},
	{"~/spectrum-1GHz-8GHz-100kHz", true, {1.0e9f, 8.0e9f, 100.0e3f}, 100},
	{"~/iq-100MHz-10MHz", false, {100.0e6f, 10.0e6f}, 1000},
	{"~/iq-1GHz-40MHz", false, {1.0e9f, 40.0e6f}, 100},
#endif
};


// Macros
//-----------------------------------------------------------------------------
#define _arraysize(a)	(sizeof(a) / sizeof(a[0]))


// CaptureToFile capture example code
// This example code opens a ThinkRF device and performs either a test sequence
// of captures or a single capture specified on the stripped command-line
// passed in. A command line of the form:
// -iq fcentre ifbw filename, or
// -spectrum fmin fmax rbw filename
// ...or blank is accepted.
//-----------------------------------------------------------------------------
bool _captureSpectrumToFile(trfHandle hDevice, _float32 fMinHz, _float32 fMaxHz, _float32 fRBWHz, const char *pFilename, _uint32 uFrames=100) {
	std::cout << "Capturing spectrum <" << fMinHz << ", " << fMaxHz << ", " << fRBWHz << "> to file " << pFilename << " for " << uFrames << " frames" << std::endl;

	// Create stream
	_float32 fReferenceLeveldBm(-50.0f);
	const char *pWindow("nuttall");
	trfHandle hStream(trfInvalidHandle);
	trfStatus eStatus = trfCreateSpectrumStream(&hStream, fMinHz, fMaxHz, fRBWHz, pWindow, fReferenceLeveldBm);
	if (eStatus != trfOk) {
		std::cout << "trfCreateSpectrumStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}
	

	// Direct output to also go to file
	_removeOutputFiles(pFilename);
	//! [trfSetStreamOutputFile]
	//!
	trfSetStreamOutputFile(hStream, pFilename);
	if (eStatus != trfOk) {
		std::cout << "trfSetStreamOutputFile failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		trfClose(&hStream);
		return false;
	}
	//! [trfSetStreamOutputFile]

	// Start capture
	eStatus = trfAttachStreamToDevice(hDevice, hStream);
	if (eStatus != trfOk) {
		std::cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		trfClose(&hStream);
		return false;
	}

	// Capture a series of frames
	for (_uint32 iIteration = 0; iIteration < uFrames; iIteration++) {
		trfSpectrumFrame *pSpectrumFrame(nullptr);
		do {
			eStatus = trfGetNextSpectrum(hStream, &pSpectrumFrame, 2000);
		} while ((eStatus == trfWaiting) || (eStatus == trfNotStarted));

		if (eStatus != trfOk) {
			std::cout << "Spectrum streaming failed at iteration " << iIteration << std::endl;
			break;
		}

		// Report and dispose of the spectrum frame
		char cDigit('0' + (pSpectrumFrame->uSequenceNumber % 10));
		std::cout << cDigit << std::flush;
		trfFreeSpectrum(&pSpectrumFrame);
	}
	std::cout << std::endl;

	// Close the stream - will wait for file end and also will perform the detach.
	trfClose(&hStream);

	return true;
}

bool _captureIQToFile(trfHandle hDevice, _float32 fCentreHz, _float32 fIFBWHz, const char *pFilename, _uint32 uFrames=100) {
	std::cout << "Capturing IQ <" << fCentreHz << ", " << fIFBWHz << "> to file " << pFilename << " for " << uFrames << " frames" << std::endl;
	// IQ capture
	trfHandle hStream(trfInvalidHandle);
	_float32 fReferenceLeveldBm(-50.0f);
	trfStatus eStatus = trfCreateIQStream(&hStream, fCentreHz, fIFBWHz, fReferenceLeveldBm);
	if (eStatus != trfOk) {
		std::cout << "trfCreateIQStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}

	// Direct output to also go to file
	eStatus = trfSetStreamOutputFile(hStream, pFilename);
	if (eStatus != trfOk) {
		std::cout << "trfSetStreamOutputFile failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}

	// Attach IQ stream to device - starts streaming. 
	eStatus = trfAttachStreamToDevice(hDevice, hStream);
	if (eStatus != trfOk) {
		std::cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		trfClose(&hStream);
		return false;
	}

	// Capture until stream reports exhausted
	// Capture a series of frames
	for (_uint32 iIteration = 0; iIteration < uFrames; iIteration++) {
		trfIQFrame *pIQFrame(nullptr);
		eStatus = trfGetNextIQ(hStream, &pIQFrame, 5000);	// Obtain next frame - specifying 5000msec timeout
		if (eStatus == trfExhausted) {
			break;	// End of capture - stream has detached
		}
		if (eStatus != trfOk) {
			std::cout << "IQ streaming failed at iteration " << iIteration << std::endl;
			break;
		}

		// Report and dispose of the spectrum frame
		char cDigit('0' + (pIQFrame->uSequenceNumber % 10));
		std::cout << cDigit << std::flush;
		trfFreeIQ(&pIQFrame);
	}
	std::cout << std::endl;

	// Dispose of this stream
	trfClose(&hStream);
	return true;
}

static bool _playbackSpectrum(const std::string &sFile) {
	//! [trfAttachStreamToFile]
	//!
	trfHandle hStream(trfInvalidHandle);
	std::cout << "Playing back Spectra from file " << sFile << std::endl;
	trfStatus eStatus(trfCreateSpectrumStream(&hStream, 0.0f, 0.0f, 0.0f, nullptr, 0.0f));
	if (eStatus != trfOk) {
		std::cout << "trfCreateSpectrumStream failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}
	eStatus = trfAttachStreamToFile(hStream, sFile.c_str());
	if (eStatus != trfOk) {
		std::cout << "trfAttachStreamToFile failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}

	char *pJSON(nullptr);
	trfGetParameters(hStream, &pJSON);
	std::cout << "Stream parameters:" << std::endl << "---" << std::endl << pJSON << std::endl;
	//! [trfAttachStreamToFile]

	while (true) {
		trfSpectrumFrame *pSpectrumFrame(nullptr);
		eStatus = trfGetNextSpectrum(hStream, &pSpectrumFrame, 1000);
		if (eStatus == trfWaiting) {
			std::cout << std::endl << "trfGetNextSpectrum returns trfWaiting..." << endl;
			continue;
		}
		else if (eStatus == trfExhausted) {
			std::cout << std::endl << "trfGetNextSpectrum returns trfExhausted" << endl;
			break;
		}
		else if (eStatus != trfOk) {
			std::cout << std::endl << "trfGetNextSpectrum failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
			break;
		}

		// Report and dispose of the spectrum frame
		std::cout << "." << std::flush;
		trfFreeSpectrum(&pSpectrumFrame);
	}
	std::cout << std::endl;
	return true;
}

static bool _playbackIQ(const std::string &sFile, bool bLooping=false) {
	char *pJSON(nullptr);
	
	trfHandle hStream(trfInvalidHandle);
	std::cout << "Playing back IQ from file " << sFile << std::endl;
	trfStatus eStatus(trfCreateIQStream(&hStream, 0.0f, 0.0f, 0.0f));
	if (eStatus != trfOk) {
		std::cout << "trfCreateIQStream failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}
	
	if (true == bLooping) {
		trfAddBooleanParameter(&pJSON, TRFLooping, true);
		eStatus = trfSetParameters(hStream, &pJSON);
		if (eStatus != trfOk) {
			std::cout << "trfSetParameters failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
			return false;
		}
	}

	eStatus = trfAttachStreamToFile(hStream, sFile.c_str());
	if (eStatus != trfOk) {
		std::cout << "trfAttachStreamToFile failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}

	//! [trfGetParameters]
	//!
	trfGetParameters(hStream, &pJSON);
	std::cout << "Stream parameters:" << std::endl << "---" << std::endl << pJSON << std::endl;
	//! [trfGetParameters]

	while (true) {
		trfIQFrame *pIQFrame(nullptr);
		eStatus = trfGetNextIQ(hStream, &pIQFrame, 1000);
		if (eStatus == trfWaiting) {
			std::cout << std::endl << "trfGetNextIQ returns trfWaiting..." << endl;
			continue;
		}
		else if (eStatus == trfExhausted) {
			std::cout << std::endl << "trfGetNextIQ returns trfExhausted" << endl;
			break;
		}
		else if (eStatus != trfOk) {
			std::cout << std::endl << "trfGetNextIQ failed:" << eStatus << ": " << trfGetTextForStatus(eStatus) << std::endl;
			break;
		}

		// Report and dispose of the spectrum frame
		std::cout << "." << std::flush;
		trfFreeIQ(&pIQFrame);
	}
	std::cout << std::endl;
	return true;
}


bool CaptureToFile(int argc, const char *argv[]) {
	std::cout << "IQ/Spectrum capture to file example" << std::endl;

	const char *pFilename(nullptr);
	_float32 fCentreHz(0.0f), fIFBWHz(0.0f);
	_float32 fMinHz(0.0f), fMaxHz(0.0f), fRBWHz(0.0f);
	_uint32 uFrames(100);
	bool bSpectrumNotIQ(false);
	if (argc != 0) {
		if (0 == strcmp(argv[0], "-iq")) {
			if (argc < 4) {
				std::cout << "\tusage: -ip <fcentre> <ifbw> <filename> [<frames>]" << std::endl;
				return false;
			}
			fCentreHz = strtof(argv[1], nullptr);
			fIFBWHz = strtof(argv[2], nullptr);
			pFilename = argv[3];
			if (argc >= 5) {
				uFrames = stoi(argv[4]);
			}
			bSpectrumNotIQ = false;
		}
		else if (0 == strcmp(argv[0], "-spectrum")) {
			if (argc < 5) {
				std::cout << "\tusage: -spectrum <fmin> <fmax> <rbw> <filename> [<frames>]" << std::endl;
				return false;
			}
			fMinHz = strtof(argv[1], nullptr);
			fMaxHz = strtof(argv[2], nullptr);
			fRBWHz = strtof(argv[3], nullptr);
			pFilename = argv[4];
			if (argc >= 6) {
				uFrames = stoi(argv[5]);
			}			
			bSpectrumNotIQ = true;
		}
	}

	// Discover any accessible ThinkRF devices and open the first one enumerated
	trfHandle hDevice(_openZerothDevice());
	if (trfInvalidHandle == hDevice) {
		std::cout << "_openZerothDevice failed" << std::endl;
		return false;
	}

	if (pFilename != nullptr) {
		// User-directed capture
		if (true == bSpectrumNotIQ) {
			_captureSpectrumToFile(hDevice, fMinHz, fMaxHz, fRBWHz, pFilename, uFrames);
		}
		else {
			_captureIQToFile(hDevice, fCentreHz, fIFBWHz, pFilename, uFrames);
		}
	} else {
		// Loop over specifications to write files
		for (_uint32 i = 0; i < _arraysize(spTests); i++) {
			std::string sFile(_goodFilename(spTests[i].pFilename));
			_removeOutputFiles(sFile.c_str());
			if (true == spTests[i].bSpectrumNotIQ) {
				_captureSpectrumToFile(hDevice, spTests[i].data.spectrum.fMinHz, spTests[i].data.spectrum.fMaxHz, spTests[i].data.spectrum.fRBWHz, sFile.c_str(), spTests[i].uFrames);
			}
			else {
				_captureIQToFile(hDevice, spTests[i].data.iq.fCentreHz, spTests[i].data.iq.fIFBWHz, sFile.c_str(), spTests[i].uFrames);
			}
		}

		// Loop over specifications to read back files
		cout << endl << "File playback tests" << endl << "-------------------------------------------" << endl;
		for (_uint32 i = 0; i < _arraysize(spTests); i++) {
			std::string sFile(_goodFilename(spTests[i].pFilename));
			if (true == spTests[i].bSpectrumNotIQ) {
				_playbackSpectrum(sFile);
			}
			else {
				_playbackIQ(sFile);
			}
		}
	}

	// Close the device handle
	trfClose(&hDevice);

	std::cout << "End of file capture example" << std::endl << "---" << std::endl << std::endl;
	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	std::cout << "libtrf example code " << __FILE__ << std::endl << "---" << std::endl;
	std::cout << "library version " << trfGetVersion() << std::endl << std::endl;

	CaptureToFile(argc, argv);

	std::cout << "Example complete" << std::endl;
}
#endif // _MAIN_