/**
* @example DemodFromFileExample.cpp
*
* Demonstrates the streaming of data from file through a
* demodulating proessor.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/

// -example DemodFromFile AM C:\tmp\AMDemod\IQOttawa_1030kHz_100kHz C:\tmp\AMDemod\AMOut

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>
#include <cstring>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"


// Processor example code
// This example code creates an IQ stream from a file, attaches a demodulator
// to it (AM, FM) and records the output audio stream to a file.
//-----------------------------------------------------------------------------
bool DemodFromFileExample(int argc, const char *argv[]) {
	cout << "DemodFromFileExample" << endl;

	// Paths are required...
	if (argc < 3) {
		cout << "\tERROR: Demodulation type, Input file path and output file path are required." << endl;
		return false;
	}
	const char *pInputFile(argv[1]), *pOutputFile(argv[2]);

	// Interpret type
	bool bFM(0 == strcmp(argv[0], "FM")), bAM(0 == strcmp(argv[0], "AM"));
	if ((false == bFM) && (false == bAM)) {
		cout << "\tERROR: Demodulation type must be one of AM, FM." << endl;
	}

	// Create processor of desired type
	trfHandle hProcessor(trfInvalidHandle);
	trfStatus eStatus(trfOk);

	eStatus = trfCreateProcessor(&hProcessor, (true==bFM)?"FMDemodulator":"AMDemodulator");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Set processor output (audio) sample rate
	char *pParameters(nullptr);
	trfAddNumericParameter(&pParameters, "OutputSampleRate", 16.0e3f);//  Program output sample rate - 16kHz
	eStatus = trfSetParameters(hProcessor, &pParameters);
	if (eStatus != trfOk) {
		cout << "trfSetParameters failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	trfDisposeString(&pParameters);

	// Create output stream
	trfHandle hAudioOut(trfInvalidHandle);
	eStatus = trfGetProcessorOutputStream(hProcessor, &hAudioOut);
	if (eStatus != trfOk) {
		cout << "trfGetProcessorOutputStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Attach the output stream to a file
	eStatus = trfSetStreamOutputFile(hAudioOut, pOutputFile);
	if (eStatus != trfOk) {
		cout << "trfSetStreamOutputFile failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Open source file IQ stream
	trfHandle hIQStream(trfInvalidHandle);
	eStatus = trfCreateIQStream(&hIQStream, 0.0f, 0.0f, 0.0f);
	eStatus = trfAttachStreamToFile(hIQStream, pInputFile);
	if (eStatus != trfOk) {
		cout << "\tERROR: trfAttachStreamToFile cannot open file <" << pInputFile << ">" << endl;
		return false;
	}

	// Attach processor to source stream
	eStatus = trfAttachProcessorToStream(hProcessor, hIQStream);
	if (eStatus != trfOk) {
		cout << "trfAttachProcessorToStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Continue until exhaustion reported
	_uint32 uRxd(0);
	trfAllowDiscarding(hIQStream, true);	// Prevents hIQStream from gating processing
	bool bExhausted(false);
	while (false == bExhausted) {
		trfBasebandFrame *pBasebandFrame(nullptr);
		eStatus = trfGetNextBaseband(hAudioOut, &pBasebandFrame, 200);
		bExhausted = (trfExhausted == eStatus);
		if (eStatus < 0) {
			cout << "trfGetNextBaseband failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			break;
		}

		if (pBasebandFrame != nullptr) {
			uRxd += pBasebandFrame->uSamples;
			trfFreeBaseband(&pBasebandFrame);
			cout << ".";
		}
		else {
			cout << "?";
		}
	}
	cout << endl;
	eStatus = trfSetStreamOutputFile(hAudioOut, nullptr);

	trfGetParameters(hProcessor, &pParameters);
	cout << "Parameters : " << endl << pParameters << endl;
	trfDisposeString(&pParameters);

	eStatus = trfDetachProcessor(hProcessor);
	if (eStatus != trfOk) {
		cout << "Whups! " << eStatus << endl;
	}
	trfDetachStream(hIQStream);

	trfClose(&hProcessor);
	trfClose(&hIQStream);

	cout << "Converting " << pOutputFile << " file pair to .wav format." << endl;
	std::string pWAVFile(std::string(pOutputFile) + ".wav");
	eStatus = trfConvertBasebandFileToWavFile(pOutputFile, pWAVFile.c_str());

	cout << "Completed. Exiting." << endl << endl;
	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	DemodFromFileExample(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
