/**
* @example DeviceEnumerationExample.cpp
*
* Tests discovery, opening, closing, re-opening of ThinkRF devices.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/


#include <iostream>
#include <iomanip>
#include <random>
#include <functional>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"


// DeviceEnumerationTest code - no adaption
//-----------------------------------------------------------------------------
bool DeviceEnumerationTest(int argc, const char *pArgv[]) {
	cout << "DeviceEnumerationTest code" << endl;

	trfStatus eStatus(trfOk);

	// Display library version code
	//! [trfGetVersion]
	//!
	char *pVersion(trfGetVersion());
	cout << pVersion << endl;
	trfDisposeString(&pVersion);
	//! [trfGetVersion]

	// Add a specific device address to the search list
	//! [trfAddDeviceAddress]
	//!
	eStatus = trfAddDeviceAddress("10.126.110.114");
	if (eStatus != trfOk) {
		cout << "trfAddDeviceAddress failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfAddDeviceAddress]

	// Discover any accessible ThinkRF devices (LAN) and open the first one enumerated
	//! [trfEnumerateDevices]
	//!
	_uint uDevices(0);
	eStatus = trfEnumerateDevices(&uDevices, true);
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}
	//! [trfEnumerateDevices]

	// Open all device handles
	trfHandle pHandles[20];
	char *pJSON(nullptr);
	_uint32 uDevicesOpened(0);
	for (_uint32 i = 0; i < uDevices; i++) {
		//! [trfExamineDevice]
		//!
		eStatus = trfExamineDevice(i, &pJSON);
		if (eStatus != trfOk) {
			cout << "trfExamineDevice [ " << i << "] failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			continue;
		}
		cout << "Device:" << i << " information: " << endl << "---" << endl << pJSON << endl;
		trfDisposeString(&pJSON);
		//! [trfExamineDevice]
	}

	// Attempt to open all devices
	for (_uint32 i = 0; i < uDevices; i++) {
		//! [trfOpenDevice]
		//!
		eStatus = trfOpenDevice(i, &pHandles[i]);
		if (eStatus != trfOk) {
			cout << "trfOpenDevice [" << i << "] failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			continue;
		}
		//! [trfOpenDevice]
		uDevicesOpened++;
	}
	cout << "Opened " << uDevicesOpened << " libtrf devices" << endl;

	// Close all open device handles
	_uint32 uClosed(0);
	for (_uint32 i = 0; i < uDevices; i++) {
		if (pHandles[i] != trfInvalidHandle) {
			//! [trfClose]
			//!
			eStatus = trfClose(&pHandles[i]);
			if (eStatus != trfOk) {
				cout << "trfClose [ " << i << "] failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				continue;
			}
			//! [trfClose]
			uClosed++;
		}
	}
	cout << "Closed " << uClosed << " libtrf devices" << endl;

	// Attempt to re-open all devices
	cout << "Attempting to re-open all devices" << endl;
	_uint32 uReOpened(0);
	for (_uint32 i = 0; i < uDevices; i++) {
		eStatus = trfOpenDevice(i, &pHandles[i]);
		if (eStatus != trfOk) {
			cout << "trfOpenDevice [" << i << "] failed: " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			continue;
		}
		uReOpened++;
	}
	cout << "Reopened " << uReOpened << " libtrf devices" << endl;

	for (_uint32 i = 0; i < uDevices; i++) {
		trfHandle hHandle(pHandles[i]);
		//! [trfResetDeviceConnection]
		//!
		eStatus = trfResetDeviceConnection(hHandle);
		if (eStatus != trfOk) {
			cout << "trfResetDeviceConnection [" << i << "] failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		}
		//! [trfResetDeviceConnection]
	}

	// Close all open device handles
	for (_uint32 i = 0; i < uDevices; i++) {
		if (pHandles[i] != trfInvalidHandle) {
			eStatus = trfClose(&pHandles[i]);
			if (eStatus != trfOk) {
				cout << "trfClose [" << i << "] failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				continue;
			}
		}
	}
	cout << "All devices closed" << endl;

	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	DeviceEnumerationTest(argc, argv);

	cout << "Example complete" << endl; 
}
#endif // _MAIN_
