/**
* @example DynamicSwitchIQCaptureExample.cpp
*
* An example that demonstrates capturing IQ sequences whilst changing
* parameters on-the-fly.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/


#include <iostream>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#define _sleep(ms)	Sleep(ms)
#else
#include <unistd.h>
#define _sleep(ms)	usleep(ms*1000)
#endif

// Include libtrf header file
#include "libtrf.h"


// Macros
//-----------------------------------------------------------------------------
#define _arraysize(a)	(sizeof(a) / sizeof(a[0]))


// IQ Capture to a stream, dynamically changing stream parameters on-the-fly.
//-----------------------------------------------------------------------------
_TRFAPI char *trfGetVerboseInternalErrorState(_uint32 uNumErrors);
bool DynamicSwitchIQCapture(int argc, const char *argv[]) {
	cout << "Dynamic switch IQ capture example" << endl;

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	char *pDeviceInfo(nullptr);
	trfExamineDevice(0, &pDeviceInfo);
	cout << "trfExamineDevice pre-open returns: " << pDeviceInfo << endl;

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	trfExamineDevice(0, &pDeviceInfo);
	cout << "trfExamineDevice post-open returns: " << pDeviceInfo << endl;

	char *pJSON(nullptr);
	trfGetParameters(hDevice, &pJSON);
	std::cout << "Device parameters: " << std::endl << pJSON << endl;
	trfDisposeString(&pJSON);

	// Set of tests to run
	const static struct {
		_float32 fCentreHz;
		_float32 fIFBWHz;
		_uint32 uFramesBeforeSwitch;
	}spTests[] = {
		{1.0e9f, 1.0e6f, 200},
		{2.0e9f, 1.0e6f, 38},
		{3.0e9f, 1.0e6f, 171},
		{4.0e9f, 10.0e6f, 49},
		{5.0e9f, 20.0e6f, 211},
		{6.0e9f, 40.0e6f, 12},
		{7.0e9f, 40.0e6f, 48},
		{6.5e9f, 40.0e6f, 93},
	};

	// Run tests and report. Demonstrates reprogramming an IQ stream. Could equally close this one
	// and create a new one if needed.
	bool bFirstTest(true);
	std::cout << std::unitbuf;
	trfHandle hStream(trfInvalidHandle);
	for (_uint32 i = 0; i < _arraysize(spTests); i++) {
		cout << "Capturing IFBW " << std::scientific << spTests[i].fIFBWHz << " for " << std::fixed << spTests[i].uFramesBeforeSwitch << "frames" << endl << "----" << endl;
		
		// Re-use same stream handle (except first time)
		if (trfInvalidHandle == hStream) {
			eStatus = trfCreateIQStream(&hStream, spTests[i].fCentreHz, spTests[i].fIFBWHz, -80.0f);
			if (eStatus != trfOk) {
				cout << "trfCreateIQStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				return false;
			}
			// Attach IQ stream to device - starts streaming. 
			eStatus = trfAttachStreamToDevice(hDevice, hStream);
			if (eStatus != trfOk) {
				cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				trfClose(&hStream);
				trfClose(&hDevice);
				return false;
			}
		}
		else {
			// Reprogram running stream
			trfAddNumericParameter(&pJSON, TRFFCentreHz, spTests[i].fCentreHz);
			trfAddNumericParameter(&pJSON, TRFIFBWHz, spTests[i].fIFBWHz);
			eStatus = trfSetParameters(hStream, &pJSON);
			trfDisposeString(&pJSON);
			if (eStatus != trfOk) {
				trfClose(&hStream);
				trfClose(&hDevice);
				cout << "trfSetParameters failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				trfStatus eStatus(trfOk);
				while ((eStatus = trfGetNextStatus()) != trfOk) {
					cout << "\terror code: " << (int)eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				}
				return false;
			}
		}

		// Report initial Stream parameters (retrieved as JSON string)
		trfGetParameters(hStream, &pJSON);
		cout << "Stream parameters before capture start:" << endl << pJSON << endl;
		trfDisposeString(&pJSON);

		// Capture until stream reports exhausted
		_uint32 uSamplesCaptured(0);
		for(_uint32 uFrame = 0; uFrame < spTests[i].uFramesBeforeSwitch; uFrame++) {
			trfIQFrame *pIQFrame(nullptr);
			eStatus = trfGetNextIQ(hStream, &pIQFrame, 5000);	// Obtain next frame - specifying 5sec timeout
			if (eStatus == trfExhausted) {
				break;	// End of capture - stream has detached
			}
			if (eStatus != trfOk) {
				cout << "trfGetNextIQ failed with code (possibly timed-out) " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				const char *pError(trfGetVerboseInternalErrorState(100));
				cout << "trfGetVerboseInternalErrorState() => " << endl << "---" << endl << pError << endl << endl;
				trfClose(&hStream);
				trfClose(&hDevice);
				return false;
			}

			// Report the received frame
			uSamplesCaptured += (_uint32)pIQFrame->uSamples;
			trfFreeIQ(&pIQFrame);
			cout << (uFrame % 10);
		}
		cout << endl << "----" << endl << endl;
	}

	// Dispose of this stream
	trfClose(&hStream);

	// Close the stream and device handles
	trfClose(&hDevice);

	cout << "End of Dynamic switching IQ capture example" << endl << "---" << endl << endl;
	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	DynamicSwitchIQCapture(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
