/**
* @example FMDemodulationExample.cpp
*
* Demonstrates the setup of the FMDemodulator processor type and its usage.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"

#undef _TempFile
#undef _DefaultFile
#ifdef _WIN32
#define _TempFile		"C:\\tmp\\libtrf-audioout"
#define _DefaultFile	"C:\\tmp\\AMOut.wav"
#else
#define _TempFile		"~/libtrf-audioout"
#define _DefaultFile	"~/tmp/AMOut.wav"
#endif


// Processor example code
// This example code opens a ThinkRF device, creates a Spectrum Stream and
// a SpectrumPower processor. The processor is attached to the stream which
// is attached to the device. The example code displays the output from the
// processor for a few seconds before tearing down the structure.
//-----------------------------------------------------------------------------
bool FMDemodulationExample(int argc, const char *argv[]) {
	cout << "FMDemodulationExample" << endl;

	// Interpret target frequency
	_float32 fCentreHz(99.7e6f), fIFBWHz(200.0e3f);
	if (argc >= 2) {
		fCentreHz = strtof(argv[0], nullptr);
		fIFBWHz = strtof(argv[1], nullptr);
		cout << "Interpreted FCentre: " << fCentreHz << ", IFBW: " << fIFBWHz << endl;
	}

	// Interpret target filename
	const char *pOutputFile(_DefaultFile);
	if (argc >= 3) {
		pOutputFile = argv[2];
	}

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Grab FM station (Campbell River)
	trfHandle hIQStream(trfInvalidHandle);
	eStatus = trfCreateIQStream(&hIQStream, fCentreHz, fIFBWHz, -80.0f);
	if (eStatus != trfOk) {
		cout << "trfCreateIQStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Get the list of available processor types
	char *pTypes(nullptr);
	eStatus = trfGetProcessorTypes(&pTypes);
	cout << "trfGetProcessorTypes returns: " << endl << pTypes << endl;

	trfHandle hProcessor(trfInvalidHandle);
	eStatus = trfCreateProcessor(&hProcessor, "FMDemodulator");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Display power meter parameter set
	char *pParameterInfo(nullptr), *pParameters(nullptr);
	eStatus = trfGetParameterInfo(hProcessor, &pParameterInfo);
	cout << "Parameter info : " << endl << pParameterInfo << endl;

	
	// Set parameters
	trfAddNumericParameter(&pParameters, "OutputSampleRate", 44.1e3f);//  Program output sample rate
	//trfAddTextParameter(&pParameters, "type", "DSB");// Program AM demodulator type
	eStatus = trfSetParameters(hProcessor, &pParameters);
	if (eStatus != trfOk) {
		cout << "trfSetParameters failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	trfDisposeString(&pParameters);

	// Create output stream
	trfHandle hAudioOut(trfInvalidHandle);
	eStatus = trfGetProcessorOutputStream(hProcessor, &hAudioOut);

	// Attach the elements into a processing chain
	trfAttachStreamToDevice(hDevice, hIQStream);
	trfAttachProcessorToStream(hProcessor, hIQStream);

	// Attach the output stream to a file
	remove(_TempFile);
	eStatus = trfSetStreamOutputFile(hAudioOut, _TempFile);

	// Loop displaying frames output for 1k frames
	for (int i = 0; i < 100; i++) {
		trfBasebandFrame *pFrame(nullptr);
		while (true) {
			eStatus = trfGetNextBaseband(hAudioOut, &pFrame, 1000);
			if (eStatus == trfOk) {
				break;
			}
		}

		if (pFrame->pJSONInfo != nullptr) {
			cout << "Annotation JSON: " << pFrame->pJSONInfo << endl;
		}		
		cout << "Frame " << i << " @" << pFrame->iTimestampNanosec << endl;
	}
	eStatus = trfSetStreamOutputFile(hAudioOut, nullptr);

	trfGetParameters(hProcessor, &pParameters);
	cout << "Parameters : " << endl << pParameters << endl;
	trfDisposeString(&pParameters);

	eStatus = trfDetachProcessor(hProcessor);
	if (eStatus != trfOk) {
		cout << "Whups! " << eStatus << endl;
	}
	trfDetachStream(hIQStream);

	trfClose(&hAudioOut);
	trfClose(&hProcessor);
	trfClose(&hIQStream);
	trfClose(&hDevice);

	eStatus = trfConvertBasebandFileToWavFile(_TempFile, pOutputFile);

	return true;
}


//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	FMDemodulationExample(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
