/**
* @example FiniteIQCaptureExample.cpp
*
* An example that demonstrates capturing finite time sequences of IQ data
* under the control of libtrf including setup and synchronization.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/


#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>
#include <math.h>
using namespace std;

// Include libtrf header file
#include "libtrf.h"

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#define _sleep(ms)	Sleep(ms)
#else
#include <unistd.h>
#define _sleep(ms)	usleep(ms*1000)
#endif

std::string _generateFilename(_float32 fFrequencyMHz, _float32 fIFBWMHz, _float32 fDurationSec) {
	// Create filename string
	std::stringstream ss;
	ss << "c:\\tmp\\IQ_" << std::fixed << std::setprecision(0) << fFrequencyMHz << "MHz_"
		<< std::fixed << std::setprecision(0) << fIFBWMHz << "MHz_"
		<< std::fixed << std::setprecision(0) << (fDurationSec * 1.0e3f) << "msec.csv";
	std::string filename(ss.str());

	return filename;
}


// Macros
//-----------------------------------------------------------------------------
#define _arraysize(a)	(sizeof(a) / sizeof(a[0]))


// IQ Finite-time capture example code
// This example code opens a ThinkRF device and performs a series of IQ
// block captures, reporting the characteristics of the captures. Each capture
// runs for a pre-programmed period and self-terminates (note that a finite
// capture IQ stream self-detaches from its source device). Also demonstrates
// adding an additional programming parameter to the stream before attaching
// to a capture device.
//-----------------------------------------------------------------------------
bool FiniteIQCapture(int argc, const char *argv[]) {
	cout << "IQ Finite time capture example" << endl;

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	char *pDeviceInfo(nullptr);
	trfExamineDevice(0, &pDeviceInfo);
	cout << "trfExamineDevice pre-open returns: " << pDeviceInfo << endl;

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	trfExamineDevice(0, &pDeviceInfo);
	cout << "trfExamineDevice post-open returns: " << pDeviceInfo << endl;

	char *pJSON(nullptr);
	trfGetParameters(hDevice, &pJSON);
	std::cout << "Device parameters: " << std::endl << std::string(pJSON) << endl;
	trfDisposeString(&pJSON);

	// Set of tests to run
	const static struct {
		_float32 fCentreHz;
		_float32 fIFBWHz;
		_float32 fDurationSec;
	}spTests[] = {
		{1.00e9f, 40.0e6f, 0.05f},	// API2-167
		/*
		{1.0e9f, 1.0e6f, 0.000001f},	// Attempt to force API2-76
		{2.0e9f, 1.0e6f, 5.0f},			// 1GHz, 1MHz IFBW, 5 seconds
		{3.0e9f, 1.0e6f, 1.0f},
		{4.0e9f, 10.0e6f, 3.2f},
		{5.0e9f, 20.0e6f, 1.0f},
		{6.0e9f, 40.0e6f, 1.0f},
		{7.0e9f, 40.0e6f, 0.2f},
		{6.5e9f, 40.0e6f, 0.2f},
		*/
	};

	// Run tests and report. Demonstrates reprogramming an IQ stream. Could equally close this one
	// and create a new one if needed.
	bool bFirstTest(true);
	std::cout << std::unitbuf;
	trfHandle hStream(trfInvalidHandle);
	for (_uint32 i = 0; i < _arraysize(spTests); i++) {
		cout << "Capturing IFBW " << std::scientific << spTests[i].fIFBWHz << " for " << std::fixed << spTests[i].fDurationSec << "sec" << endl << "----" << endl;
		
		// Re-use same stream handle (except first time)
		if (trfInvalidHandle == hStream) {
			eStatus = trfCreateIQStream(&hStream, spTests[i].fCentreHz, spTests[i].fIFBWHz, -20.0f);
			if (eStatus != trfOk) {
				cout << "trfCreateIQStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				return false;
			}
		}

		// Add duration parameter (and any other reprogramming from above)
		trfAddNumericParameter(&pJSON, TRFDurationSec, spTests[i].fDurationSec);
		trfAddNumericParameter(&pJSON, TRFFCentreHz, spTests[i].fCentreHz);
		trfAddNumericParameter(&pJSON, TRFIFBWHz, spTests[i].fIFBWHz);
		trfAddNumericParameter(&pJSON, TRFBufferSec, spTests[i].fDurationSec);
		eStatus = trfSetParameters(hStream, &pJSON);
		trfDisposeString(&pJSON);
		if (eStatus != trfOk) {
			trfClose(&hStream);
			trfClose(&hDevice);
			cout << "trfSetParameters failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			trfStatus eStatus(trfOk);
			while ((eStatus = trfGetNextStatus()) != trfOk) {
				cout << "\terror code: " << (int)eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			}
			return false;
		}

		// Attach IQ stream to device - starts streaming. 
		eStatus = trfAttachStreamToDevice(hDevice, hStream);
		if (eStatus != trfOk) {
			cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			trfClose(&hStream);
			trfClose(&hDevice);
			return false;
		}

		// Report all IQ Stream parameters (retrieved as JSON string)
		trfGetParameters(hStream, &pJSON);
		cout << "Stream parameters before capture start:" << endl << pJSON << endl;
		trfDisposeString(&pJSON);

		// Capture until stream reports exhausted
		_uint32 uFramesCaptured(0), uSamplesCaptured(0);
		std::string sFilename(_generateFilename(spTests[i].fCentreHz / 1.0e6f, spTests[i].fIFBWHz / 1.0e6f, spTests[i].fDurationSec));
		std::ofstream cIQFile(sFilename);
		while (true) {
			trfIQFrame *pIQFrame(nullptr);
			eStatus = trfGetNextIQ(hStream, &pIQFrame, 5000);	// Obtain next frame - specifying 5sec timeout
			if (eStatus == trfExhausted) {
				break;	// End of capture - stream has detached
			}
			if (eStatus != trfOk) {
				cout << "trfGetNextIQ failed with code (possibly timed-out) " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				const char *pError(trfGetDetailedStatus());
				cout << "trfGetDetailedStatus() => " << endl << "---" << endl << pError << endl << endl;
				trfClose(&hStream);
				trfClose(&hDevice);
				return false;
			}

			// Report the received frame
			if (0 == uFramesCaptured) {
				_float32 fExpectedFrames(0);
				trfGetParameters(hStream, &pJSON);
				trfReadNumericParameter(&pJSON, TRFFramesExpected, &fExpectedFrames);
				cout << "Frames expected: " << (_uint32)floorf(fExpectedFrames) << endl;
			}
			uFramesCaptured++;
			uSamplesCaptured += (_uint32)pIQFrame->uSamples;

			// Output the samples to file
			const _complex32 *pcData(pIQFrame->pData);
			for (_uint32 i = 0; i < pIQFrame->uSamples; i++) {
				cIQFile << pcData[i].fI << ", " << pcData[i].fQ << endl;
			}

			trfFreeIQ(&pIQFrame);
			cout << ".";
		}
		cIQFile.close();

		cout << endl << uFramesCaptured << " frames captured - total " << uSamplesCaptured << " samples" << endl;

		//! [trfIsAttached]
		//!
		if (trfIsAttached(hStream) != trfDetached) {
			cout << "Stream is still attached to source" << endl;
			return false;
		}
		//! [trfIsAttached]

		// Report all IQ Stream parameters (retrieved as JSON string)
		trfGetParameters(hStream, &pJSON);
		cout << "Stream parameters post-capture:" << endl << pJSON << endl << endl;
		trfDisposeString(&pJSON);

		if (true == bFirstTest) {
			bFirstTest = false;
			// For the first test - redo the capture, but pull no data - just wait for the detach to occur
			trfAttachStreamToDevice(hDevice, hStream);
			
			// Wait for completion and automatic detach
			cout << "Streaming and waiting for detach to occur" << endl;
			while (trfIsAttached(hStream) != trfDetached) {
				cout << "#";
				_sleep(100);
			}
			cout << endl;

			// Grab 1st frame (to prove it exists)
			trfIQFrame *pIQFrame(nullptr);
			eStatus = trfGetNextIQ(hStream, &pIQFrame, 1000);
			if (eStatus != trfOk) {
				cout << "No frames captured/buffered " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
				trfClose(&hStream);
				trfClose(&hDevice);
				return false;
			}
			trfFreeIQ(&pIQFrame);

			// Flush stream
			eStatus = trfFlushStream(hStream);
			eStatus = trfGetNextIQ(hStream, &pIQFrame, 1000);
			cout << "Flushed stream and trfGetNextIQ returns " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		}

		cout << "----" << endl << endl;
	}

	// Dispose of this stream
	trfClose(&hStream);

	// Close the stream and device handles
	trfClose(&hDevice);

	cout << "End of IQ Finite capture example" << endl << "---" << endl << endl;
	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	FiniteIQCapture(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
