/**
* @example IQCaptureExample.cpp
*
* Opens a ThinkRF device and initiates IQ Capture, reporting the data
* retrieved.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/


#include <iostream>
#include <iomanip>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"


// IQ Streaming example code - no adaption
// This example code opens a ThinkRF device and initiates IQ streaming, reporting
// the data retrieved.
//-----------------------------------------------------------------------------
bool IQCapture(int argc, const char *argv[]) {
	cout << "IQ Streaming example without reference level adaption" << endl;

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	// Test device zero
	char *pJSON(nullptr);
	eStatus = trfExamineDevice(0, &pJSON);
	if (eStatus != trfOk) {
		cout << "trfExamineDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Create an IQ data stream
	//! [trfCreateIQStream]
	//!
	_float32 fFCentreHz(2440.0e6f), fIFBWHz(40.0e6f);
	_float32 fInitialRefLeveldBm(-50.0f);
	trfHandle hStream(trfInvalidHandle);
	eStatus = trfCreateIQStream(&hStream, fFCentreHz, fIFBWHz, fInitialRefLeveldBm);
	if (eStatus != trfOk) {
		cout << "trfCreateIQStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfCreateIQStream]

	// Attach IQ stream to device - starts streaming. 
	//! [trfAttachStreamToDevice]
	//!
	eStatus = trfAttachStreamToDevice(hDevice, hStream);
	if (eStatus != trfOk) {
		cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		char *pFullError(trfGetDetailedStatus());
		cout << "error info: " << std::string(pFullError) << endl;
		trfDisposeString(&pFullError);
		trfStatus eStatus(trfOk);
		while ((eStatus = trfGetNextStatus()) != trfOk) {
			cout << "Error code " << (int)eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		}
		trfClose(&hStream);
		trfClose(&hDevice);
		return false;
	}
	//! [trfAttachStreamToDevice]

	// Report all IQ Stream parameters (retrieved as JSON string)
	trfGetParameters(hStream, &pJSON);
	cout << "Stream parameters before capture start:" << endl << pJSON << endl;
	trfDisposeString(&pJSON);

	//! [trfGetNextIQ]
	//!
	// Report the first uFramesRemaining IQFrames received and then detach
	_uint uFramesRemaining(1000);
	_uint32 uFramesCaptured(0), uSamplesCaptured(0);
	while (uFramesRemaining > 0) {
		trfIQFrame *pIQFrame(nullptr);
		eStatus = trfGetNextIQ(hStream, &pIQFrame, 5000);	// Obtain next frame
		if (eStatus != trfOk) {
			cout << "trfGetNextIQ failed with code (possibly timed-out) " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
			trfClose(&hStream);
			trfClose(&hDevice);
			return false;
		}

		// Report the received frame
		uFramesCaptured++;
		uSamplesCaptured += (_uint32)pIQFrame->uSamples;
		if (0 == (uFramesCaptured % 10)) {
			cout << uFramesCaptured << " frames captured - total " << uSamplesCaptured << " samples" << endl;
		}
		uFramesRemaining--;

		// Release the received frame
		trfFreeIQ(&pIQFrame);
	}
	//! [trfGetNextIQ]

	//! [trfDetachStream]
	//!
	// Detach stream halts capture
	trfDetachStream(hStream);
	if (trfIsAttached(hStream) != trfDetached) {
		cout << "Not showing stream as detached - weird" << endl;
		return false;
	}
	//! [trfDetachStream]

	// Detach file (if set)
	trfSetStreamOutputFile(hStream, nullptr);

	// Flush out any remaining spectra from this capture - continue until
	// trfExhausted is returned from trfGetNextSpectrum
	cout << "flushing any remaining IQ Frames";
	while (true) {
		trfIQFrame *pIQFrame(nullptr);
		eStatus = trfGetNextIQ(hStream, &pIQFrame, 100);
		if (eStatus == trfExhausted) {
			break;
		}
		cout << ".";

		// Dispose of the IQ frame
		trfFreeIQ(&pIQFrame);
	}
	cout << endl << "flushing completed" << endl;

	// Report all IQ Stream parameters (retrieved as JSON string)
	char *pPostJSON(nullptr);
	trfGetParameters(hStream, &pPostJSON);
	cout << "Stream parameters post-capture:" << endl << pPostJSON << endl;
	trfDisposeString(&pPostJSON);

	// Close the stream and device handles
	trfClose(&hStream);
	trfClose(&hDevice);

	cout << "End of IQ Streaming example" << endl << "---" << endl << endl;
	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	IQCapture(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_