/**
* @example IQToSpectrumExample.cpp
*
* Demonstrates the setup of a processor and its usage.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"

/*
void _badProgramming(trfHandle hIQStream) {
	char *pJSON(nullptr);
	trfAddNumericParameter(&pJSON, TRFRefLevel, -30);
	trfAddNumericParameter(&pJSON, TRFFCentreHz, (9000e6f + 8000000000e6f) / 2.0f);
	trfAddNumericParameter(&pJSON, TRFIFBWHz, 8000000000e6f - 9000e6f);
	trfStatus eStatus(trfSetParameters(hIQStream, &pJSON));
	if (eStatus != trfOk) {
		std::cout << std::endl << "trfSetParameters returns (" << eStatus << ") - " << trfGetTextForStatus(eStatus) << std::endl;
		std::cout << "Further info: " << trfGetDetailedStatus() << std::endl;
	}
}

void _goodProgramming(trfHandle hIQStream) {
	char *pJSON(nullptr);
	trfAddNumericParameter(&pJSON, TRFIFBWHz, 1.0e6f);
	trfStatus eStatus(trfSetParameters(hIQStream, &pJSON));
	if (eStatus != trfOk) {
		std::cout << std::endl << "trfSetParameters returns (" << eStatus << ") - " << trfGetTextForStatus(eStatus) << std::endl;
		std::cout << "Further info: " << trfGetDetailedStatus() << std::endl;
	}
}
*/

static void _reportPeak(trfSpectrumFrame *pSpectrumFrame) {
	if (nullptr == pSpectrumFrame) {
		return;
	}

	// Find peak and index
	_uint32 uPeakIndex(0);
	_float32 fPeakdBm(pSpectrumFrame->pData[0]);
	for (_uint32 i = 1; i < pSpectrumFrame->uPoints; i++) {
		if (pSpectrumFrame->pData[i] > fPeakdBm) {
			fPeakdBm = pSpectrumFrame->pData[i];
			uPeakIndex = i;
		}
	}

	// Convert to frequency
	_float32 fWidthHz((_float32)(pSpectrumFrame->fMaxHz - pSpectrumFrame->fMinHz));
	_float32 fPeakOffsetHz(fWidthHz * ((_float32)uPeakIndex / ((_float32)(pSpectrumFrame->uPoints - 1))));
	_float32 fPeakFHz((_float32)pSpectrumFrame->fMinHz + fPeakOffsetHz);

	// Report
	std::cout << "Found peak at " << fPeakFHz << "Hz at " << fPeakdBm << "dBm" << std::endl;
}


// IQToSpectrumExample code
// This example code opens a ThinkRF device, creates an IQ stream and attaches
// an IQToSpectrum processor in order to get a IQ and spectrum stream
// From the IQToSpectrum processor, the spectrum stream is then chained
// with the spectrum characaterization processor to get trace data. 
// This demonstrates  chaining of  multiple processors togather.
// IQ stream → IQToSpectrum Processor → Spectrum stream → 
// SpectrumCharacterization Processor → spectrum stream with trace data 
// (min, max, average-hold spectra). User can extend this example by getting
// the output of the final stream and then chaining it with another processor. 
//-----------------------------------------------------------------------------
bool IQToSpectrumExample(int argc, const char *argv[]) {
	cout << "IQToSpectrum processor example" << endl;
	if (argc > 1) {
		trfAddDeviceAddress(argv[1]);	// Andrew's unit
	}

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed with code " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed with code " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Setup IQ stream
	trfHandle hIQStream(trfInvalidHandle);
	_float32 fCentreHz(1.0e9f), fIFBWHz(40.0e6f);
	eStatus = trfCreateIQStream(&hIQStream, fCentreHz, fIFBWHz, -40.0f);
	cout << "Created IQ Stream centred on " << fCentreHz << "Hz, requested " << fIFBWHz << "Hz IFBW" << endl;
	if (eStatus != trfOk) {
		cout << "trfCreateIQStream failed with code " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	// Get the list of available processor types
	char *pTypes(nullptr);
	eStatus = trfGetProcessorTypes(&pTypes);
	cout << "trfGetProcessorTypes returns: " << endl << pTypes << endl;

	// Create Processor for getting spectrum stream.
	trfHandle hProcessor(trfInvalidHandle);
	eStatus = trfCreateProcessor(&hProcessor, "IQToSpectrum");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed with code " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Display processor parameter set
	char *pParameterInfo(nullptr), *pParameters(nullptr);
	eStatus = trfGetParameterInfo(hProcessor, &pParameterInfo);
	cout << "Parameter info : " << endl << pParameterInfo << endl;

	// Set processor input parameters
	trfAddBooleanParameter(&pParameters, TRFFollowIQ, true);	
	trfAddTextParameter(&pParameters, TRFWindow, "flattop");
	eStatus = trfSetParameters(hProcessor, &pParameters);
	//trfAddNumericParameter(&pParameters, "RBWHz", 1.0e3f);		// 1KHz RBW
	//trfAddNumericParameter(&pParameters, "FFTOverlap", 0.5f);	// 50% overlap
	// Follow-IQ - overrides RBW and FFTOverlap and guarentees one spectrum frame for every IQ frame. 
	if (eStatus != trfOk) {
		cout << "trfSetParameters failed with code " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		while ((eStatus == trfGetNextStatus()) != trfOk) {
			cout << " - adnl: " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		}
		return false;
	}

	// Set IQ stream additional parameters
	trfAddBooleanParameter(&pParameters, TRFFlowControl, true);
	eStatus = trfSetParameters(hIQStream, &pParameters);
	if (eStatus != trfOk) {
		cout << "trfSetParameters failed with code " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		while ((eStatus == trfGetNextStatus()) != trfOk) {
			cout << " - adnl: " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		}
		return false;
	}

	// Create output spectrum stream.
	trfHandle hIQSpectrumOut(trfInvalidHandle);
	eStatus = trfGetProcessorOutputStream(hProcessor, &hIQSpectrumOut);

	// Attach the elements into a processing chain
	trfAttachStreamToDevice(hDevice, hIQStream);
	trfAttachProcessorToStream(hProcessor, hIQStream);

	// we now want to attach this output spectrum stream to the spectrum characterization stream. First we create it.
	trfHandle hSpectrumChar(trfInvalidHandle);
	std::cout << "creating processor for: " << "SpectrumChar" << std::endl;
	eStatus = trfCreateProcessor(&hSpectrumChar, "SpectrumCharacterization");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed with code:: " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Display SpectrumChar parameter set
	char *pSpectrumCharParameterInfo(nullptr), *pSpectrumCharParamters(nullptr);
	eStatus = trfGetParameterInfo(hSpectrumChar, &pSpectrumCharParameterInfo);
	cout << "SpectrumCharacterizationProcessor info : " << endl << pSpectrumCharParameterInfo << endl;

	trfAddNumericParameter(&pSpectrumCharParamters, "Average", 100);
	trfSetParameters(hSpectrumChar, &pSpectrumCharParamters);
	trfDisposeString(&pSpectrumCharParamters);

	// Display programmed parameters
	trfGetParameters(hSpectrumChar, &pSpectrumCharParamters);
	cout << "Spectrum Characterization parameters : " << endl << pSpectrumCharParamters << endl;
	trfDisposeString(&pSpectrumCharParamters);
	trfHandle hSpecStream(trfInvalidHandle);

	// Attach the elements into a processing chain
	trfAttachProcessorToStream(hSpectrumChar, hIQSpectrumOut);
	trfHandle hSpectrumOut(trfInvalidHandle);
	eStatus = trfGetProcessorOutputStream(hSpectrumChar, &hSpectrumOut);

	// Get and report metadata for first IQ frame.
	bool bFirstIQ(true), bFirstSpectrum(true);
	_uint32 uIQFrames(0), uSpectrumFrames(0);
	std::cout << setprecision(10);
	
	// keep getting frames until we get 1000 frames then exit.
	while(uSpectrumFrames < 100) {
		trfIQFrame *pIQFrame(nullptr);
		eStatus = trfGetNextIQ(hIQStream, &pIQFrame, 1000);
		if (trfOk == eStatus) 
		{
			//std::cout << "Found IQ Frame: " << pIQFrame->uSequenceNumber << std::endl;
			trfFreeIQ(&pIQFrame);
			uIQFrames++;
		} 
		else 
		{
			std::cout << "Failed to get IQ frame with err = " << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
			continue;
		}

		// Handle spectra for this IQ frame. 
		trfSpectrumFrame *pSpectrumFrame(nullptr);
		eStatus = trfGetNextSpectrum(hSpectrumOut, &pSpectrumFrame, 1000);
		if (trfOk == eStatus) {
			_reportPeak(pSpectrumFrame);
			//std::cout << "Found Spectrum Frame[" << uSpectrumFrames << "]: " << pSpectrumFrame->uSequenceNumber << " with number of traces: " << pSpectrumFrame->uAdditionalSpectra << std::endl;
			trfFreeSpectrum(&pSpectrumFrame);
			uSpectrumFrames++;
		}
		else
		{
			std::cout << "Failed to get spectrum frame with err = " << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
			continue;
		}
	}
	
	trfGetParameters(hProcessor, &pParameters);
	std::cout << "Processor parameters : " << std::endl << pParameters << std::endl;
	trfDisposeString(&pParameters);

	/// Cleanup. 
	//---------------------------------
	// detach stream. 
	eStatus = trfDetachStream(hIQStream);
	std::cout << "Detached stream with status = " << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;

	// detach processors. 
	eStatus = trfDetachProcessor(hProcessor);
	if (eStatus != trfOk) {
		std::cout << "Whups! IQToSpectrum processor " << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
	}
	
	// detach processors. 
	eStatus = trfDetachProcessor(hSpectrumChar);
	if (eStatus != trfOk) {
		std::cout << "Whups! spec characterization processor " << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
	}

	trfClose(&hProcessor);
	trfClose(&hSpectrumChar);
	trfClose(&hIQStream);
	trfClose(&hDevice);

	return true;
}


//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	IQToSpectrumExample(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
