/**
* @example SCProcessorExample.cpp
*
* Run spectrum characterization processor on spectrum stream from an RTSA device.
*
* Language: C++11
* aly.khedr@thinkrf.com 11/11/2022
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"


// FinteCaptureSpectrumCharacterizationProcessorExample
// In this example the SpectrumCharacterizationProcessor Processor consumes 
// a stream of Spectrum information and generates trace measurements over 
// that stream. These measurements include max-hold spectrum, min-hold spectrum, 
// and frame averaged spectrum. In addition, this example demostrates the use
// of finite capture. In finite capture mode, every user selected number of frames,
// the frame will reset and and a new set of frames is received from the device. Every 
// set of frames the stream previous stream is exhausted and must be reattached. 
//-----------------------------------------------------------------------------
#define kCaptureFrames		(10)
#define kNewCaptureLimit		(100)
bool FiniteCaptureSpectrumCharacterizationProcessorExample(int argc, const char *argv[]) {
	cout << "Processor example" << endl;
	if (argc > 1) {
		trfAddDeviceAddress(argv[1]);	// Andrew's unit
	}

	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return false;
	}

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Sweep over user selected settings
	_float32 fStartHz(0.009e6);
	_float32 fStopHz(27000.000512e6);
	_float32 fRBWHz(488.281e3);
	trfHandle hSpectrumStream(trfInvalidHandle);

	eStatus = trfCreateSpectrumStream(&hSpectrumStream, fStartHz, fStopHz, fRBWHz, "nuttall", -40.0f);
	if (eStatus != trfOk) {
		cout << "trfCreateSpectrumStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	char* pJSON(nullptr);

	// Get the list of available processor types
	char *pTypes(nullptr);
	eStatus = trfGetProcessorTypes(&pTypes);
	cout << "trfGetProcessorTypes returns: " << endl << pTypes << endl;

	// Create a SpectrumChar processor
	trfHandle hSpectrumChar(trfInvalidHandle);
	eStatus = trfCreateProcessor(&hSpectrumChar, "SpectrumCharacterization");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed with code:: " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// Display SpectrumChar parameter set
	char *pSpectrumCharParameterInfo(nullptr), *pSpectrumCharParamters(nullptr);
	eStatus = trfGetParameterInfo(hSpectrumChar, &pSpectrumCharParameterInfo);
	cout << "SpectrumCharacterizationProcessor info : " << endl << pSpectrumCharParameterInfo << endl;

	// Program processor with inputs. 
	trfAddNumericParameter(&pSpectrumCharParamters, TRFAverageCount, 100);

	// Set pre-buffer size for hSpectrumChar processor
	trfAddNumericParameter(&pSpectrumCharParamters, TRFBufferFrames, kCaptureFrames);
	trfSetParameters(hSpectrumChar, &pSpectrumCharParamters);
	trfDisposeString(&pSpectrumCharParamters);

	// Display programmed parameters
	trfGetParameters(hSpectrumChar, &pSpectrumCharParamters);
	cout << "Spectrum Characterization parameters : " << endl << pSpectrumCharParamters << endl;
	trfDisposeString(&pSpectrumCharParamters);

	// Create output stream that will contain the spectrum and measurements data.
	trfHandle hSpectrumOut(trfInvalidHandle);
	trfGetProcessorOutputStream(hSpectrumChar, &hSpectrumOut);
	trfAddNumericParameter(&pJSON, TRFBufferFrames, kCaptureFrames);
	trfSetParameters(hSpectrumOut, &pJSON);

	// Attach the elements into a processing chain
	trfAttachStreamToDevice(hDevice, hSpectrumStream);
	trfAttachProcessorToStream(hSpectrumChar, hSpectrumStream);

	// kCaptureFrames frames x 5 times. 
	_uint frames(0);
	_uint prevIndex(0);
	_uint iter(0);
	bool bFirstFrame(true);
	while (1) {
		trfSpectrumFrame* pFrame(nullptr);
		prevIndex = ((kCaptureFrames - 1) == prevIndex) ? 0 : prevIndex;
		eStatus = trfGetNextSpectrum(hSpectrumOut, &pFrame, 1000);
		if (eStatus == trfExhausted)
		{	
			std::cout << "TRFEXHAUSTED ............................................." << std::endl;
			break;
		}
		else if (eStatus == trfWaiting)
		{
			std::cout << "I am waiting:" << std::endl;
			continue;
		}
		else if (eStatus == trfOk)
		{
	
			std::cout << "Got Spectrum: " << pFrame->uSequenceNumber << std::endl;
 
			trfFreeSpectrum(&pFrame);
		}
		else
		{
			std::cout << "Some Error: " <<  eStatus << std::endl;
			continue;
		}
		frames++;		
		if (frames == 10)
		{
	        trfAddNumericParameter(&pJSON, TRFRefLevel, -30.0f);
			trfAddNumericParameter(&pJSON, TRFFMinHz, 13000.004756e6f);
			trfAddNumericParameter(&pJSON, TRFFMaxHz, 14000.004756e6f);
			trfAddNumericParameter(&pJSON, TRFRBWHz, 200e3f);
			trfAddNumericParameter(&pJSON, TRFDurationFrames, kNewCaptureLimit);
			trfAddTextParameter(&pJSON, TRFWindow, "nuttall");	
			eStatus = trfSetParameters(hSpectrumStream, &pJSON);
			if (eStatus != trfOk)
			{
				std::cout << "Failed to set spectrum parameters." << std::endl;
				std::cout << trfGetDetailedStatus() << std::endl;
				return eStatus;
			}	
		}
		iter++;
	}
	trfDetachProcessor(hSpectrumChar);
	trfDetachStream(hSpectrumStream);
	trfClose(&hSpectrumChar);
	trfClose(&hSpectrumStream);
	trfClose(&hDevice);

	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	SpectrumCharacterizationProcessorExample(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
