/**
* @example SCProcessorFromFIleExample.cpp
*
* Run spectrum characterization processor on spectrum stream from a previously
* recorded spectrum file.
*
* Language: C++11
* aly.khedr@thinkrf.com 11/11/2022
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif
#include <filesystem>

#ifdef __linux
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#endif

#include "libtrf.h"

#ifndef _MAIN_

#define _arraysize(a)	(sizeof(a) / sizeof(a[0]))

trfHandle _openZerothDevice(void) {
	// Discover any accessible ThinkRF devices and open the first one enumerated
	_uint uDevices(0);
	trfStatus eStatus(trfEnumerateDevices(&uDevices, true));
	if (eStatus != trfOk) {
		cout << "trfEnumerateDevices failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	if (0 == uDevices) {
		cout << "No ThinkRF devices enumerated" << endl;
		return trfInvalidHandle;
	}

	// Examine device (before opening)
	char *pJSON(nullptr);
	trfExamineDevice(0, &pJSON);
	cout << "trfExamineDevice" << endl << "---" << endl;
	cout << pJSON << endl << endl;

	// Open a handle to the zeroth device
	trfHandle hDevice(trfInvalidHandle);
	eStatus = trfOpenDevice(0, &hDevice);
	if (eStatus != trfOk) {
		cout << "trfOpenDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return trfInvalidHandle;
	}
	trfGetParameterInfo(hDevice, &pJSON);
	cout << "trfGetParameterInfo" << endl << "---" << endl;
	cout << pJSON << endl << endl;

	trfGetParameters(hDevice, &pJSON);
	cout << "trfGetParameters" << endl << "---" << endl;
	cout << pJSON << endl << endl;
	trfDisposeString(&pJSON);

	trfAddTextParameter(&pJSON, TRFPLLSource, "internal");
	eStatus = trfSetParameters(hDevice, &pJSON);

	return hDevice;
}

std::string _goodFilename(std::string sSourceName) {
#ifdef _WIN32
	return sSourceName;
#else
	const char *pHomeDir(getenv("HOME"));
	if (nullptr == pHomeDir) {
		pHomeDir = getpwuid(getuid())->pw_dir;
	}
	std::string sHomeDir(pHomeDir);
	if (sSourceName[0] = '~') {
		sSourceName = sHomeDir + sSourceName.substr(1);
	}
	return sSourceName;
#endif
}

static void _removeOutputFiles(const char *pRootFilename) {
	char pFullFilename[256];
	sprintf(pFullFilename, "%s.json", pRootFilename);
	remove(pFullFilename);
	sprintf(pFullFilename, "%s.data", pRootFilename);
	remove(pFullFilename);
}

// CaptureToFile capture example code
// This example code opens a ThinkRF device and performs either a test sequence
// of captures or a single capture specified on the stripped command-line
// passed in. A command line of the form:
// -iq fcentre ifbw filename, or
// -spectrum fmin fmax rbw filename
// ...or blank is accepted.
//-----------------------------------------------------------------------------
bool _captureSpectrumToFile(trfHandle hDevice, _float32 fMinHz, _float32 fMaxHz, _float32 fRBWHz, const char *pFilename, _uint32 uFrames = 100) {
	std::cout << "Capturing spectrum <" << fMinHz << ", " << fMaxHz << ", " << fRBWHz << "> to file " << pFilename << " for " << uFrames << " frames" << std::endl;

	// Create stream
	_float32 fReferenceLeveldBm(-50.0f);
	const char *pWindow("nuttall");
	trfHandle hStream(trfInvalidHandle);
	trfStatus eStatus = trfCreateSpectrumStream(&hStream, fMinHz, fMaxHz, fRBWHz, pWindow, fReferenceLeveldBm);
	if (eStatus != trfOk) {
		std::cout << "trfCreateSpectrumStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		return false;
	}

	//char *pJSON(nullptr);
	//trfAddNumericParameter(&pJSON, "BufferFrames", (_float32)uFrames);
	//trfSetParameters(hStream, &pJSON);

	// Direct output to also go to file
	_removeOutputFiles(pFilename);
	trfSetStreamOutputFile(hStream, pFilename);

	// Start capture
	eStatus = trfAttachStreamToDevice(hDevice, hStream);
	if (eStatus != trfOk) {
		std::cout << "trfAttachStreamToDevice failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << std::endl;
		trfClose(&hStream);
		return false;
	}

	// Capture a series of frames
	for (_uint32 iIteration = 0; iIteration < uFrames; iIteration++) {
		trfSpectrumFrame *pSpectrumFrame(nullptr);
		do {
			eStatus = trfGetNextSpectrum(hStream, &pSpectrumFrame, 2000);
		} while ((eStatus == trfWaiting) || (eStatus == trfNotStarted));

		if (eStatus != trfOk) {
			std::cout << "Spectrum streaming failed at iteration " << iIteration << std::endl;
			break;
		}

		// Report and dispose of the spectrum frame
		char cDigit('0' + (pSpectrumFrame->uSequenceNumber % 10));
		std::cout << cDigit;
		trfFreeSpectrum(&pSpectrumFrame);
	}
	std::cout << std::endl;

	// Close the stream - will wait for file end and also will perform the detach.
	trfClose(&hStream);

	return true;
}
#endif // ifdef _MAIN_


#ifdef _WIN32
const char *spTestFileSpec("C:\\tmp\\spectrumtest");
#else
const char *spTestFileSpec("~/spectrumtest");
#endif


// SpectrumCharacterizationProcessorExample
// This example code opens a ThinkRF device, creates a source spectrum stream from file and attached to a
//  SpectrumCharacterizationProcessor to get measurements data from the spectrum data coming out of a file.
//-----------------------------------------------------------------------------
bool SpectrumCharacterizationProcessorExampleFromFile(int argc, const char *argv[]) {
	// Capture an example spectrum sequence to file and clean up
	trfHandle hDevice(_openZerothDevice());
	if (trfInvalidHandle == hDevice) {
		return false;
	}
	
	// Create capture and save to file.
	_captureSpectrumToFile(hDevice, 87.5e6f, 108.0e6f, 1.0e3, _goodFilename(spTestFileSpec).c_str(), 100);

	// Close the device
	trfClose(&hDevice);

	// 1. Get the list of available processor types
	char *pTypes(nullptr);
	trfStatus eStatus(trfGetProcessorTypes(&pTypes));
	cout << "trfGetProcessorTypes returns: " << endl << pTypes << endl;

	// 2. Create a SpectrumChar processor
	trfHandle hSpectrumChar(trfInvalidHandle);
	std::cout <<"creating processor for: " << "SpectrumChar" << std::endl;
	eStatus = trfCreateProcessor(&hSpectrumChar, "SpectrumCharacterization");
	if (eStatus != trfOk) {
		cout << "trfCreateProcessor failed with code:: " << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	// 3. Display SpectrumChar parameter set
	char *pSpectrumCharParameterInfo(nullptr), *pSpectrumCharParamters(nullptr);
	eStatus = trfGetParameterInfo(hSpectrumChar, &pSpectrumCharParameterInfo);

	// 4. Display programmed parameters
	trfGetParameters(hSpectrumChar, &pSpectrumCharParamters);
	cout << "Spectrum Characterization parameters : " << endl << pSpectrumCharParamters << endl;
	trfDisposeString(&pSpectrumCharParamters);

	// 5. Get processor output stream handle
	trfHandle hSpectrumOut(trfInvalidHandle);
	eStatus = trfGetProcessorOutputStream(hSpectrumChar, &hSpectrumOut);

	// 6. Create source spectrum stream (will attach to file) and set parameters
	trfHandle hSpectrumStream(trfInvalidHandle);
	trfCreateSpectrumStream(&hSpectrumStream, 0.0f, 0.0f, 0.0f, nullptr, 0.0f);
	trfAllowDiscarding(hSpectrumStream, true);	// Stream not serviced - prevents blocking
	char *pJSON(nullptr);
	trfAddBooleanParameter(&pJSON, TRFLooping, true);
	trfSetParameters(hSpectrumStream, &pJSON);

	// 7. Attach the processor to hSpectrumStream
	eStatus = trfAttachProcessorToStream(hSpectrumChar, hSpectrumStream);
	if (eStatus != trfOk) {
		return false;
	}

	// 8. Now processing chain is built, create/start the input stream
	eStatus = trfAttachStreamToFile(hSpectrumStream, _goodFilename(spTestFileSpec).c_str());
	if (eStatus != trfOk) {
		return false;
	}

	// Loop displaying frames output for uFramesToGo frames and get processor output associated with frame.
	std::cout << "Starting stream from file..." << std::endl;
	char *pStreamParams(nullptr);
	_uint32 uFrameNumber(0), uFramesToGo(1031);
	while (true) {
		trfSpectrumFrame *pFrame(nullptr);
		eStatus = trfGetNextSpectrum(hSpectrumOut, &pFrame, 5000);
		if (eStatus == trfExhausted) {
			std::cout << "trfExhausted" << std::endl;
			break;
		}
		else if (eStatus == trfWaiting) {
			std::cout << "trfWaiting..." << std::endl;
			continue;
		}
		else if (eStatus == trfNotStarted) {
			std::cout << "trfNotStarted" << std::endl;
			continue;
		}
		else if (eStatus == trfOk) {
			uFrameNumber++;
			uFramesToGo = (uFramesToGo > 0) ? (uFramesToGo - 1) : 0;
			std::cout << "Got spectrum from file: " << pFrame->uSequenceNumber << ", uFramesToGo: " << uFramesToGo << ", traces: " << pFrame->uAdditionalSpectra << std::endl;
			trfFreeSpectrum(&pFrame);

			// Force break at end of sequence (at end of next run through file)
			if (0 == uFramesToGo) {
				char *pJSON(nullptr);
				trfAddBooleanParameter(&pJSON, TRFLooping, false);
				trfSetParameters(hSpectrumStream, &pJSON);
				//trfDetachStream(hSpectrumStream);		// For immediate stop, otherwise finishes after pass-through of file
			}	
		}
		else {
			std::cout << "Unexpected trfGetNextSpectrum return code " << eStatus << ", " << trfGetTextForStatus(eStatus) << endl;
			break;
		}
	}

	// Cleanup.
	trfDetachProcessor(hSpectrumChar);
	trfDetachStream(hSpectrumStream);
	trfClose(&hSpectrumChar);
	trfClose(&hSpectrumStream);
	trfAllowDiscarding(hSpectrumStream, false);

	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	SpectrumCharacterizationProcessorExampleFromFile(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
