/**
* @example SpectrumCaptureExample.cpp
*
* This example code opens a ThinkRF device and initiates Spectrum streaming, reporting
* the data retrieved.
*
* Language: C++11
* Code Copyright (C) 2022 ThinkRF Incorporated. All Rights Reserved.
*/


#include <iostream>
#include <sstream>
#include <fstream>
using namespace std;

// Cross platform standard utilities
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "libtrf.h"
#include <vector>
#include <list>

_TRFAPI char *trfGetVerboseInternalErrorState(_uint32 uMaxReportErrors);

void timeDetach(trfHandle hStream) {
	auto cStart(std::chrono::high_resolution_clock::now());
	while (trfIsAttached(hStream) != trfDetached) {
		std::this_thread::sleep_for(std::chrono::milliseconds(1));
	}
	auto cStop(std::chrono::high_resolution_clock::now());
	auto cDuration(std::chrono::duration_cast<std::chrono::milliseconds>(cStop - cStart));
	std::cout << "delay to detach = " << cDuration.count() << " milliseconds" << std::endl;
}


// This example code opens a ThinkRF device and spectrum capture, reporting
// the data retrieved. 
//-----------------------------------------------------------------------------
bool SpectrumCapture(int argc, const char *argv[]) {
	trfStatus eStatus(trfOk);
	char *pJSON(nullptr);

	cout << "Spectrum Capture example" << endl;

	trfHandle hDevice(_openZerothDevice());
	if (trfInvalidHandle == hDevice) {
		return false;
	}

	//! [trfCreateSpectrumStream]
	//!
	// Create a spectrum (sweep) stream from 1 to 2GHz with RBW of 20kHz
	cout << "Creating a spectrum stream." << endl;
	_float32 fFStartHz(1000.0e6f), fFStopHz(2000.0e6f), fRBWHz(20.0e3f), fRefLeveldBm(-50.0f);
	const char *pWindow(nullptr);
	trfHandle hStream(trfInvalidHandle);
	eStatus = trfCreateSpectrumStream(&hStream, fFStartHz, fFStopHz, fRBWHz, pWindow, fRefLeveldBm);
	if (eStatus != trfOk) {
		cout << "trfCreateSpectrumStream failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}
	//! [trfCreateSpectrumStream]

	// Add flow-control flag
	trfAddBooleanParameter(&pJSON, TRFFlowControl, true);
	eStatus = trfSetParameters(hStream, &pJSON);
	if (eStatus != trfOk) {
		cout << "trfSetParameters failed:" << eStatus << " - " << trfGetTextForStatus(eStatus) << endl;
		return false;
	}

	std::list<_float32> lFCentreMHz({ 50.0, 100.0f, 200.0f, 500.0f, 1000.0f, 2000.0f, 3000.0f, 5000.0f, 8000.0f });
	for (auto cFCMHz = lFCentreMHz.begin(); cFCMHz != lFCentreMHz.end(); ++cFCMHz) {
		std::cout << "Hit <return> when ready with signal at " << (*cFCMHz) << "MHz" << std::endl;
		getchar();

		for (fRBWHz = 1000.0f; fRBWHz > 50.0f; fRBWHz /= 2) {
			cout << "Setting RBW to " << fRBWHz << "Hz" << endl;
			trfAddNumericParameter(&pJSON, TRFRBWHz, fRBWHz);
			trfAddNumericParameter(&pJSON, TRFDurationFrames, 12);	// 12 frames only
			trfSetParameters(hStream, &pJSON);

			// Start capture
			if (trfDetached == trfIsAttached(hStream)) {	// Start capture
				eStatus = trfAttachStreamToDevice(hDevice, hStream);
			}

			// Capture a series of frames
			//! [trfGetNextSpectrum]
			//!
			_uint32 uPoints(0);
			_float32 fRBWHz(0.0f);
			bool bFirstFrame(true);
			for (_uint32 i = 0; i < 100; i++) {
				trfSpectrumFrame *pSpectrumFrame(nullptr);
				do {
					eStatus = trfGetNextSpectrum(hStream, &pSpectrumFrame, 5000);
					if (eStatus < trfOk) {
						cout << "trfGetNextSpectrum failed: " << trfGetTextForStatus(eStatus) << ":" << (int)eStatus << endl;
						break;
					}
					else if (eStatus == trfNotStarted) {
						cout << "s";
						std::cout << trfGetVerboseInternalErrorState(100) << std::endl;
						break;
					}
					else if (eStatus == trfWaiting) {
						cout << "w";
					}
					else if (eStatus == trfExhausted) {
						cout << "!";
						break;
					}
				} while (pSpectrumFrame == nullptr);

				if (nullptr == pSpectrumFrame) {
					break;
				}

				if (uPoints != pSpectrumFrame->uPoints) {
					uPoints = pSpectrumFrame->uPoints;
					cout << "Frame points " << uPoints << endl;
				}
				if (fRBWHz != pSpectrumFrame->fRBWHz) {
					fRBWHz = pSpectrumFrame->fRBWHz;
					cout << "RBWHz " << fRBWHz << endl;
				}
				trfFreeSpectrum(&pSpectrumFrame);
				cout << (i%10) << flush;
			}
			//! [trfGetNextSpectrum]
			cout << endl;
			break;
		}

		// Detach the stream - halts capture. Note that there may be unreported
		// spectrum frames still in the stream which remain accessible.
		if (trfOk == trfIsAttached(hStream)) {
			trfDetachStream(hStream);
		}
		trfFlushStream(hStream);

		// Stream parameters dump post capture
		trfGetParameters(hStream, &pJSON);
		cout << "Stream parameters post capture:" << endl << std::string(pJSON) << endl;
		trfDisposeString(&pJSON);
		break;
	}

	// Close the stream and device handles
	trfClose(&hStream);
	trfClose(&hDevice);

	// And we are done
	return true;
}

//-----------------------------------------------------------------------------
#ifndef _MAIN_
int main(int argc, const char *argv[]) {
	cout << "libtrf example code " << __FILE__ << endl << "---" << endl;
	cout << "library version " << trfGetVersion() << endl << endl;

	SpectrumCapture(argc, argv);

	cout << "Example complete" << endl;
}
#endif // _MAIN_
