/******************************************************************************
* @file libtrf.h
*
* @brief ThinkRF Device common API header file
*******************************************************************************
* Language: C99 ISO/IEC 9899:1999 
* File Version: 1.0
* @date 12Apr2023
*******************************************************************************
* Code Copyright (C) 2020-2023 ThinkRF Incorporated. All Rights Reserved.
******************************************************************************/

#ifndef _LIBTRF_
#define _LIBTRF_


//*** Macros
//*****************************************************************************
// Build-control for Windows and Linux platforms
//-----------------------------------------------------------------------------
#if defined (_WIN32) | defined (_WIN64)
 #ifdef _DLL
  #define _TRFAPI extern "C" __declspec(dllexport)
 #else
  #define _TRFAPI extern "C" __declspec(dllimport)
 #endif
#elif defined (__GNUG__)
 #include <sys/stat.h>
 #include <fcntl.h> 
 #ifdef _SHAREDLIB
  #define _TRFAPI __attribute__((visibility("default"))) extern "C"
 #else
  #define _TRFAPI extern "C"
 #endif
#else
 #error "Unknown build platform"
#endif

//*** Defines
//*****************************************************************************
/// Invalid handle value for any libtrf entity. As good practice, all unused
/// or uninitialized handle values should as good practice be set to this value - ie.
/// @code
/// trfHandle hNewDevice(trfInvalidHandle);
/// @endcode
#define trfInvalidHandle		(0xffffffff)


/// Generic JSON parameter names for libtrf entities; devices, streams, processors.
/// These parameter names are used in conjunction with JSON string construction/
/// examination functions to allow parameters to be set for and obtained from
/// libtrf entities. In the following documentation 'RO' denotes 'read-only', and
/// RW denotes read/write. Setting a parameter for an entity that does not support
/// that parameter is an error and will typically return trfParameterSetError when
/// calling trfSetParameters with a JSON string including an unsupported parameter.
/// Detailed information can be obtained using trfGetDetailedStatus. Note that these
/// names may change across future releases and therefore it is highly recommended
/// that the following symbolic names TRF___ are used instead of their textual
/// equivalent.
/// RO - Read-only, RW - Read and Write
//-----------------------------------------------------------------------------
#ifndef _PARAMETERS_

// JSON names used for defining parameter ranges and similar
#define TRFDefault					"default"			///< RO default parameter value tag
#define TRFMin						"min"				///< RO minimum value tag
#define TRFMax						"max"				///< RO maximum value tag
#define TRFType						"type"				///< RO type of unit tag (user readable)
#define TRFUnits					"units"				///< RO units of value tag
#define TRFAddress					"address"			///< RO network address tag

// Entity-specific JSON parameter names
#define TRFDevice					"device"			///< RO device information section
#define TRFDeviceType				"type"				///< RO parameter - device textual type
#define TRFDeviceVersion			"version"			///< RO parameter - device textual version
#define TRFDeviceSerialNum			"serial"			///< RO parameter - device textual serial number
#define TRFDeviceFirmware			"firmware"			///< RO parameter - device textual firmware version
#define TRFDeviceConnection			"connection"		///< RO parameter - device textual connection information
#define TRFHasFlatteningInfo		"iqequalization"	///< RO receiver flag that equalization data is available (flattening is possible)

#define TRFSequence					"Sequence"			///< RO sequence number parameter (internal)
#define TRFTimestamp				"Timestamp"			///< RO timestamp parameter (internal)
#define TRFTimeResolution			"TimeResolution"	///< RO time resolution parameter (internal)
#define TRFBandwidthHz				"BandwidthHz"		///< RO generic bandwidth parameter

#define TRFFilename					"filename"			///< RW parameter - filename parameter (stream)
#define TRFLooping					"LoopingFlag"		///< RW set to cause file to loop
#define TRFSweepActive				"SweepActive"		///< RO parameter - sweep is active flag (Rx)
#define TRFIQActive					"IQActive"			///< RO parameter - IQ capture is active flag (Rx)
#define TRFBufferSec				"bufferSec"			///< RW parameter - user-settable maximum buffering duration in seconds.
														///< If set this will override TRFBufferFrames.
#define TRFBufferFrames				"bufferFrames"		///< RW parameter - user-settable maximum stream buffering in frames.
														///< If set this will override TRFBufferSec.
#define TRFDiscardProportion		"bufferDiscard"		///< RW parameter - On overflow (TRFBufferSec, TRFBufferFrames) this sets
														///< the proportion of the buffer that will be discarded. The oldest frames
														///< will be discarded first. Setting to zero will minimize the number of
														///< frames discarded. Setting to 1.0 will discard the whole buffer. All other
														///< values 0,1 are acceptable. This allows a user to manage the tradeoff between
														///< discontinuities and data latency.
#define TRFFMinHz					"FMinHz"			///< RW parameter - Min extent of sweep range
#define TRFFMaxHz					"FMaxHz"			///< RW parameter - Max extent of sweep range
#define TRFRBWHz					"RBWHz"				///< RW parameter - Resolution Bandwidth in Hz.
#define TRFWindow					"WindowFn"			///< RW parameter - Window function for spectrum reconstruction
#define TRFFCentreHz				"FCentreHz"			///< RW parameter - IQ Capture centre frequency
#define TRFSampleRateHz				"SampleRateHz"		///< RO parameter - IQ Capture sample rate
#define TRFIFBWHz					"IFBWHz"			///< RW parameter - IF Bandwidth for IQ capture.
														///< Note that in the current implementation of libtrf, IF bandwidths narrower
														///< than 100kHz are not supported. Additionally IF bandwidths of less than 1MHz
														///< below 50MHz FCentre exhibit significant distortion and their use is not
														///< recommended. This limitation will be addressed in planned future releases.
#define TRFRefLevel					"RefdBm"			///< RW parameter - User-supplied expected max signal dBm
#define TRFUserCalibrationOffset	"UserCaldB"			///< RW parameter - User-supplied correction-factor for both Spectrum and calibrated-IQ
#define TRFFlattenFlag				"flatten"			///< RW parameter - apply spectrum equalization if available
#define TRFDurationFrames			"captureFrames"		///< RW parameter - Spectrum Stream specified capture duration
#define TRFDurationSec				"captureSec"		///< RW parameter - IQ Stream specified capture duration
#define TRFMaxContiguousSec			"maxContiguousSec"	///< RO parameter - IQ Stream max limit of continuous streaming. Note -only valid after attaching the stream to a receiver device.
#define TRFIsContiguous				"contiguous"		///< RO parameter - Indicates this IQ stream is guaranteed contiguous.
#define TRFFramesExpected			"framesExpected"	///< RO parameter - Number of frames expected in this stream, or 0 if configured for continuous capture.
#define TRFSubOptimalFlag			"SubOptimalDR"		///< RO parameter - Dynamic range is non-optimal flag
#define TRFFullAdaptFlag			"adaptFull"			///< RW parameter - Full adapt flag - if set on stream will cause DR adaption before IQ streaming (~500msec)
#define TRFStepAdaptFlag			"adaptStep"			///< RW parameter - Step adapt flag - if set on stream will cause DR adaption before IQ streaming
#define TRFMeasuredAvgdBm			"avgdBm"			///< RO parameter - Measured average power - only valid if TRFRefLevelAdapt is true
#define TRFdBFSD					"dBFSD"				///< RO parameter - Proportion of dynamic range utilized - expressed in dB relative to full-scale
#define TRFFlowControl				"flowControl"		///< RW parameter - If set, ensures tight flow control with source if it is a receiver for spectra and IQ data

#define TRFAttenuation				"attenuation"		///< RO parameter - Attenuation used in dB
#define TRFPacketDurationMsec		"PacketDuration"	///< RO parameter - Packet duration in secondds
#define TRFFilterRatio				"IFFilterRatio"		///< RO parameter - IF Filter ratio applied
#define TRFNTP						"NTPServers"		///< RW parameter - NTP servers list
#define TRFSCPIQueryTimeout			"SCPIQueryTimeout"	///< RW parameter - SCPI transaction timeout

#define TRFRxSampleRate				"RxSampleRate"		///< RO parameter - native 'base' sample rate of a receiver
#define TRFGNSSValid				"GNSSValid"			///< RO parameter - true if GNSS data is valid, false otherwise
#define TRFLatitude					"latitude"			///< RO parameter - Latitude returned by GNSS
#define TRFLongitude				"longitude"			///< RO parameter - Longitude returned by GNSS
#define TRFAltitude					"altitude"			///< RO parameter - Altitude returned by GNSS
#define TRFGNSSRxTime				"GNSSPosnEpoch"		///< RO parameter - Last position report in UTC Msec
#define TRFGNSSTimestamp			"GNSSTimeNanosec"	///< RO parameter - Most recent GNSS UTC timestamp in nanosec
#define TRFGNSSADelay				"GNSSAntDelay"		///< RW parameter - GNSS Antenna delay parameter
#define TRFGNSSConstellation		"GNSSCons"			///< RO parameter - GNSS Constellation information
#define TRFGNSSDynamicMode			"GNSSDynamic"		///< RW parameter - GNSS Dynamic mode (textual)
#define TRFGNSSSpeedOverGround		"GNSSSpeed"			///< RO parameter - GNSS Speed over ground in m/sec
#define TRFGNSSHeadingDegT			"GNSSHeading"		///< RO parameter - GNSS Heading in degrees true
#define TRFGNSSTrackDegT			"GNSSTrack"			///< RO parameter - GNSS Track in degrees true
#define TRFGNSSMagVarDegT			"GNSSMagVar"		///< RO parameter - GNSS Magnetic variation in degrees E+,W-

#define TRFDeviceTemperature		"deviceTemperature"	///< RO parameter - Reported device internal temperatures
#define TRFDeviceLockStatus			"deviceClockLock"	///< RO parameter - Reported device clock lock statuses
#define TRFPLLSource				"pllSource"			///< RW parameter - PLL clock source (internal, external, gnss)
#define TRFPPSSource				"ppsSource"			///< RW parameter - PPS clock source (external, gnss)
#define TRFPPSSource				"ppsSource"			///< RW parameter - PPS clock source (external, gnss)
#define TRFTimeSync					"timeSync"			///< RW parameter - clock reference source (disable,"ntp,once","ntp,continuous")
#define TRFCurrentDate				"Date"				///< RW parameter - current date <year>,<month>,<date>
#define TRFCurrentTime				"Time"				///< RW parameter - current time <hour>,<minute>,<second>[,<millisecond>]

// AM/FM Demodulator parameter names
#define TRFAMSubtype				"type"				///< RW parameter - type of AM demodulation to perform; [DSB, LSB, USB]
#define TRFFramesProduced			"frames"			///< RO parameter - count of frames produced
#define TRFOutputSampleRate			"OutputSampleRate"	///< RW parameter - requested baseband output sample rate

// IQ to Spectrum processor parameter names (re-uses TRFRBWHz, TRFWindowFn)
#define TRFFollowIQ					"FollowIQ"			///< RW parameter - if 'true', then one spectrum frame will be produced for
														///< each input IQ frame with RBW dictated by the window
														///< function and input IQ frame size.
#define TRFOverlap					"FFTOverlap"		///< RW parameter - sets the proportion 0.0-1.0 of overlap in
														///< successive output spectra.
#define TRFMatchIQ					"MatchIQ"			///< RO parameter - Most recently processed input IQ frame sequence number.
#define TRFOutputSize				"OutputSize"		///< RO parameter - Size of output spectra expected with current parameters.
#define TRFRecentFrames				"RecentFrames"		///< RO parameter - Number of frames processed since last trfGetParameters call.

// Spectrum Characterization processor parameter names
#define TRFAverageCount				"Average"			///< RW parameter - number of spectra over which to create 'average' spectrum.
#define TRFClearAverageTrace		"ClearAverageTrace" ///< RW parameter - On the next frame, clear the average trace.
#define TRFClearMaxHoldTrace		"ClearMaxHoldTrace" ///< RW parameter - On the next frame, clear the max-hold trace.
#define TRFClearMinHoldTrace		"ClearMinHoldTrace" ///< RW parameter - On the next frame, clear the min-hold trace.

#endif // _PARAMETERS_


//*** Type definitions
//*****************************************************************************
/// Basic types used in the API
//-----------------------------------------------------------------------------
#ifndef _TRFTYPES_

#ifdef __GNUC__
typedef char _int8;				///< Signed 8-bit integer
typedef short _int16;			///< Signed 16-bit integer
typedef int _int32;				///< Signed 32-bit integer
typedef long long int _int64;	///< Signed 64-bit integer
#endif

typedef unsigned short _uint16;	///< Unsigned 16-bit integer
typedef unsigned int _uint32;	///< Unsigned 32-bit integer
typedef unsigned int _uint;		///< Unsigned integer
typedef unsigned long long int _uint64;	///< Signed 64-bit integer
typedef float _float32;			///< 32-bit floating point (IEEE-754) value
typedef double _float64;		///< 64-bit floating point (IEEE-754) value

/// Complex type consisting of 32-bit float Real (In-phase) and Imaginary (Quadrature) components
typedef struct {
	_float32 fI;	///< 32-bit floating point (IEEE-754) In-phase (I) sample value
	_float32 fQ;	///< 32-bit floating point (IEEE-754) Quadrature (Q) sample value
}_complex32;

#endif //_TRFTYPES_


/// Device or processor handle type
//-------------------------------------
typedef _uint32	trfHandle;	///< Handle for accessing libtrf resource (device, processor etc...)


/// General operation status reporting
//-------------------------------------
typedef enum _trfStatus {
	trfOk = 0,		///< No errors, no warnings - operation has completed successfully
	
	/// Warnings or status (greater than 0)
	trfWaiting = 1,				///< No error - data delivery is underway but momentarily unavailable
	trfExhausted = 2,			///< No more data is available
	trfNotStarted = 3,			///< Data is expected but is not yet available
	trfDiscontinuousWithPreviousFrame = 4,	///< Indicator that a discontinuity is detected in IQ or baseband data
	trfDetached = 5,			///< Indicates that a stream is in a 'detached' state
	//...

	/// Errors codes (less than 0)
	trfUnimplemented = -1,			///< Function or behaviour is not currently available
	trfAPINotInitialized = -2,		///< API initialization is required before executing this function

	trfDeviceAddressAlreadyKnown = -3,	///< An already-known device address has been added
	trfUnallocatedUserData = -4,	///< Expected user-allocated storage has not been provided
	trfDeviceIndexOutOfRange = -5,	///< Provided device index is not in the range of enumerated devices
	trfNoDeviceInformation = -6,	///< No information is available for the specified device
	trfCannotOpenDevice = -7,		///< Unable to open the specified device
	trfBadDeviceHandle = -8,		///< The given device handle is invalid or corrupted
	trfBadUnitHandle = -9,			///< The given entity handle is invalid or corrupted
	trfBadParameterName = -10,		///< A parameter name specified is incorrect
	trfNoParametersSpecified = -11,	///< No parameters where parameters were expected have been specified
	tffBadParameter = -12,			///< A parameter has been specified which is out of range
	trfInvalidParameter = -13,		///< An invalid parameter has been specified
	trfBadReceiverHandle = -14,		///< The given receiver handle is invalid or corrupted
	trfIQStreamStartFailure = -15,	///< The IQ stream has failed to start
	trfMemoryAllocationError = -16, ///< An internal memory allocation has failed
	trfInvalidStreamHandle = -17,	///< The given stream handle is invalid or corrupted
	trfNoAssociatedRxOrFile = -18,	///< The specified stream has no associated data source
	trfNotAnIQStream = -19,			///< An IQ-specific call is made on a non IQ-stream entity handle
	trfNoDataSpecified = -20,		///< No data has been specified where data was expected
	trfInvalidHandleSpecified = -21,///< The given entity handle is invalid or corrupted
	trfBadStreamShutdown = -22,		///< An error was encountered during stream shutdown
	trfUnsupportedParameter = -23,	///< A parameter was specified which is not supported
	trfCannotCloseStream = -24,		///< An error occurred preventing the specified stream being closed
	trfNotASpectrumStream = -25,	///< A stream handle for a non-spectrum stream was specified
	trfCannotRestartStream = -26,	///< A failure has occurred preventing the stream from restarting
	trfSweepStartFailure = -27,		///< A failure has occurred preventing the spectrum stream from starting
	trfFrameTypeMismatch = -28,		///< An incorrect frame type has been specified
	trfFileOpenFailed = -29,		///< A failure has occurred when attempting to open a file
	trfFileTypeIsNotIQ = -30,		///< The file specified was expected to contain IQ data
	trfFileTypeIsNotSpectrum = -31,	///< The file specified was expected to contain spectra
	trfNoAssociatedSource = -32,	///< No source is associated with this entity
	trfUnknownParameter = -33,		///< The parameter specified is not known
	trfParameterSetError = -34,		///< An error was encountered when setting a parameter
	trfDeviceHandleIsInvalid = -35,	///< An invalid or corrupted device handle was specified
	trfSCPISendFailed = -36,		///< An error in sending SCPI data has occurred
	trfSCPIQueryFailed = -37,		///< An error in making a SCPI query has occurred
	trfBadStreamHandle = -38,		///< An invalid or corrupted stream handle was specified
	trfCannotAttachIQStream = -39,		///< A failure occurred when attempting to attach an IQ stream
	trfCannotAttachSpectrumStream = -40,///< A failure occurred when attempting to attach a spectrum stream
	trfCannotAttachStream = -41,		///< A failure occurred when attempting to attach a stream
	trfCannotDetachIQStream = -42,		///< A failure occurred when attempting to detach an IQ stream
	trfCannotDetachSpectrumStream = -43,///< A failure occurred when attempting to detach a spectrum stream
	trfCannotDetachStream = -44,		///< A failure occurred when attempting to detach a stream
	trfConnectionResetFailed = -45,		///< Failed to reset the connection to the specified device
	trfBadFrequencyParameters = -46,	///< 'Impossible' frequency parameters specified (ie. FStart >= FEnd)
	trfBadRBWParameter = -47,			///< 'Impossible' RBW parameter specified
	trfBadReferenceLevelParameter = -48,///< 'Impossible' reference level parameter specified
	trfSpectrumStreamCreateFailure = -49,	///< General failure to create a spectrum stream
	trfBufferSecOutOfRange = -50,		///< Buffer seconds parameter specified is invalid
	trfBufferFramesOutOfRange = -51,	///< Buffer frames parameter specified is invalid
	trfBufferDiscardProportionOutOfRange = -52,	///< Buffer discard propotion parameter specified is out of range (valid is 0.0 to 1.0)
	trfFrequencyRangeIsInvalid = -53,	///< Invalid frequency range specified
	trfBadGNSSDynamicMode = -54,		///< Invalid GNSS dynamic mode specified
	trfBadPLLSource = -55,				///< Invalid PLL Source parameter specified
	trfPLLSourceWithNoGNSS = -56,		///< PLL source specified for a non-GNSS enabled device
	trfFCentreInvalid = -57,			///< Centre frequency specified is invalid
	trfIFBWInvalid = -58,				///< IFBW specified is invalid
	trfFMinInvalid = -59,				///< Minimum frequency specified is invalid
	trfFMaxInvalid = -60,				///< Maximum frequency specified is invalid
	trfResolutionBWInvalid = -61,		///< Resolution bandwidth specified is invalid
	trfFCentreOutOfRange = -62,			///< Centre frequency specified is out of range for this device
	trfIFBWOutOfRange = -63,			///< IFBW specified is out of range for this device
	trfReferenceLevelOutOfRange = -64,	///< Reference level is out-of-range for this device
	trfResolutionBWOutOfRange = -65,	///< Resolution bandwidth is out-of-range for this device
	trfUnknownWindowType = -66,			///< Window type specified is not supported by this device
	trfStartFrequencyOutOfRange = -67,	///< Minimum frequency specified is not supported by this device
	trfStopFrequencyOutOfRange = -68,	///< Maximum frequency specified is not supported by this device
	trfDeviceUnreachable = -69,			///< Unable to connect with device through network address
	trfInvalidFilename = -70,			///< Filename passed does not specify a valid file path
	trfCannotOpenOutputFile = -71,		///< Failed to open specified output file for writing
	trfUninitializedUserData = -72,		///< Expected data was found uninitialized
	trfBadProcessorHandle = -73,		///< Unknown processor handle specified
	trfCannotCreateProcessor = -74,		///< Failed to create processor
	trfProcessorAttachFailed = -75,		///< Attempt to attach processor to stream failed (wrong stream type?)
	trfProcessorDetachFailed = -76,		///< Attempt to detach processor from source stream failed
	trfProcessorUnattached = -77,		///< Attempt to detach processor when not attached
	trfNotABasebandStream = -78,		///< Stream handle given is not of a baseband stream
	trfFileTypeIsNotBaseband = -79,		///< Stream file given is not a baseband file
	trfBasebandStreamStartFailure = -80,///< Failed to start baseband file strema from file
	trfSampleRateInvalid = -81,			///< Invalid sample rate specified
	trfCannotObtainOutputStream = -82,	///< Unable to obtain processor output stream handle
	trfCannotConnectOutputStream = -83,	///< Failed to connect output stream to processor
	trfBadFilename = -84,				///< Filename / file path is invalid
	trfWindowTypeInvalid = -85,			///< Window type specified is not known
	trfDeviceCommunicationsLost = -86,	///< Device is not responding over network
	trfUnknownStreamType = -87,			///< Stream type is not known - internal error (should never happen)
	trfStreamHandleAlreadyOpen = -88,	///< Stream handle is already open - attempt to re-open failed.
	trfInvalidStreamSource = -89,		///< Stream source is invalid inside a connection operation
	trfBadPPSSource = -90,				///< Invalid PPS Source parameter specified

	trfLastErrorSentinel = -1000
}trfStatus;


// trfIQFrame
/// IQ data frame. This structure contains the actual data and all descriptive
/// metadata.
//-------------------------------------
typedef struct _trfIQFrame {
	trfStatus eFlags;			///< Any warning/error for this frame only.
	_uint32 uSequenceNumber;	///< Frame sequence number (monotonic unit-step)
	_int64 iTimestampNanosec;	///< Frame origination timestamp.
	_float64 fCentreHz;			///< Centre frequency
	_float32 fIFBWHz;			///< Usable IF Bandwidth - assumed symmetrical around fCentreHz
	_float32 fSampleRateHz;		///< Sample rate in Hz - fSampleRateHz is always greater than fIFBWHz
	bool bDiscontinuity;		///< 'true' if frame is not contiguous with previous frame. Will not be set for the first frame
								///< of a new stream when first attached. If detached and re-attached the first frame will have
								///< the discontinuity flag set.
	bool bSubOptimalDRFlag;		///< 'true' if a low or high excursion outside target dynamic range was detected in this frame
	_uint32 uSamples;			///< Complex IQ samples in *pData
	const _complex32 *pData;	///< IQ Frame data as calibrated sqrt(mW)
	const char *pJSONInfo;		///< Any additional information for this frame (ie. from processing)
}trfIQFrame;


// trfSpectrumFrame
/// Spectrum data frame. This structure contains the actual spectral data and all
/// descriptive metadata.
//-------------------------------------
typedef struct _trfSpectrumFrame {
	trfStatus eFlags;			///< Any warning/error for this frame only.
	_uint32 uSequenceNumber;	///< Frame sequence number (monotonic unit-step)
	_int64 iTimestampNanosec;	///< Timestamp of first sample used in spectrum reconstruction.
	_int64 iDurationNanosec;	///< Delta from first-sample to last-sample timestamp used in spectrum
								///< reconstruction. Note that this is not the same as the frame rate
								///< which can be derived from a stream by comparing the timestamps of
								///< successive trfSpectrumFrame instances in the stream.
	_float64 fMinHz;			///< Centre frequency of first bin in sweep
	_float64 fMaxHz;			///< Centre frequency of last bin in sweep
	_float32 fRBWHz;			///< Resolution Bandwidth in Hz
	_uint32 uPoints;			///< Number of data points in *pData
	const _float32 *pData;		///< Frame data as calibrated dBm
	const char *pJSONInfo;		///< Any additional information for this frame (ie. from processing)
	_uint32 uAdditionalSpectra;				///< Number of entries in pAdditionalSpectra
	const _float32 **pAdditionalSpectra;	///< Array of uAdditionalSpectra entries
}trfSpectrumFrame;


// trfBasebandFrame
/// Baseband sampled data (ie. audio) frame. Basedband data can be of any type
/// representable by an array of _float32 values. Multiple channels are
/// supported.
//-------------------------------------
typedef struct _trfBasebandFrame {
	_uint32 uSequenceNumber;	///< Frame sequence number (monotonic unit-step)
	_int64 iTimestampNanosec;	///< Frame origination timestamp.
	_float32 fSampleRateHz;		///< Sample rate in Hz
	_float32 fUsableBWHz;		///< Usable bandwidth of sampled data (is less than fSampleRateHz/2)
	_uint32 uChannels;			///< Channels in pData. Channel0 at offset 0, 1 at offset uSamples etc...
	_uint32 uSamples;			///< Samples in frame
	const _float32 **ppData;	///< ppData[0] - channel0, ppData[1] - channel 1 etc...
								///< Frame data (range +/-1.0f). Array terminates with a NULL pointer.
	const char *pJSONInfo;		///< Any additional information for this frame (ie. from processing)
}trfBasebandFrame;


//*** Global declarations
//*****************************************************************************

//*** Housekeeping functions
//*****************************************************************************
// trfGetVersion
/// Return a C-string describing this version of the library.
/// @return	textual description of the library version. 
///			Should be disposed of using trfDisposeString using delete[].
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfGetVersion
//-------------------------------------
_TRFAPI char *trfGetVersion(void);


//*** Error handling functions
//*****************************************************************************
// trfGetTextForStatus
/// Return textual (UTF-8, ISO-Latin-1) status description matching with the
/// passed status code. 
/// @param eStatus	API status code
/// @return	Error strfing - remains owned by API.
//-------------------------------------
_TRFAPI const char *trfGetTextForStatus(trfStatus eStatus);

// trfGetNextStatus
/// Return any additional (descriptive) error codes arising from the
/// last issued libtrf function on the calling thread. To return all error
/// status a user can iterate on this call until it returns trfOk to
/// indicate no further error information is available. 
/// @return		API status code (trfOk if no further status available)
//-------------------------------------
_TRFAPI trfStatus trfGetNextStatus(void);

// trfGetDetailedStatus
/// Return textual (UTF-8, ISO-Latin-1) error details for the last significant
/// status on the calling thread. Clears error string for this thread on on call.
/// @return	Error string - should be disposed of using trfDisposeString.
//-------------------------------------
_TRFAPI char *trfGetDetailedStatus(void);


//*** Device discovery, open/close
//*****************************************************************************
// trfAddDeviceAddress
/// Add a potential device address for use by trfEnumerateDevices. Note
/// that local devices that may be discovered directly (LAN, USB3 etc...) will
/// not require their addresses to be added through this call, however it is
/// not an error if they are.
/// @param pRemoteAddress IPv4/IPv6 address for remote device.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfAddDeviceAddress
//-------------------------------------
_TRFAPI trfStatus trfAddDeviceAddress(const char *pRemoteAddress);

// trfEnumerateDevices
/// Enumerate all accessible devices and return a count. Device enumeration
/// may be initiated by a user at any time to enumerate newly connected
/// devices or update the list of devices.
/// @param puDevices	  Address of _uint to receive the number of devices known.
/// @param bAutoDiscover  If 'true', then attempt to find any accessible devices.
///					   	  Otherwise only devices explicitly added through trfAddDeviceAddress
///						  are included.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfEnumerateDevices
//-------------------------------------
_TRFAPI trfStatus trfEnumerateDevices(_uint *puDevices, bool bAutoDiscover);

// trfExamineDevice
/// Return known information about a discovered device (uDeviceIndex in the
/// range [0 .. (uDevices-1)] into user-allocated trfDeviceInfo structure.
/// @param	uDeviceIndex	Index of device 
/// @param	ppJSON			Pointer to receive UTF8 JSON string
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfExamineDevice
//-------------------------------------
_TRFAPI trfStatus trfExamineDevice(_uint uDeviceIndex, char **ppJSON);

// trfOpenDevice
/// Given a device index, attempt to open it. On success sets the device
/// handle *pHandle.
/// @param uDeviceIndex	Index of device from most recent call to trfEnumerateDevices
/// @param pDevice	Pointer to device handle to be used in future operations
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfOpenDevice
//-------------------------------------
_TRFAPI trfStatus trfOpenDevice(_uint uDeviceIndex, trfHandle *pDevice);

// trfResetDeviceConnection
/// If a communications error is suspected, issuing this call will cause a connection
/// shutdown and re-open to the device without loosing any operational state. On
/// success the device will continue with programmed operation. If the return value
/// is not 'trfOk' then it is likely communications are unrecoverable and the device
/// handle should be closed (see trfClose).
/// @param cDevice		Device handle to be used in future operations
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfResetDeviceConnection
//-------------------------------------
_TRFAPI trfStatus trfResetDeviceConnection(trfHandle cDevice);

// trfClose
/// Close any open handle (device, processor, stream)
/// On successful return, *pHandle==kInvalidHandle
/// @param *pHandle		Pointer to Device, processor, stream handle
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet DeviceEnumerationExample.cpp trfClose
//-------------------------------------
_TRFAPI trfStatus trfClose(trfHandle *pHandle);


//*** Stream Handling
//*****************************************************************************
// trfCreateIQStream
/// Create an IQ stream with the given parameters. This entity will be inactive
/// until successfully attached to a device capable of servicing the stream.
/// Note that in the current implementation of libtrf, IF bandwidths narrower
/// than 100kHz are not supported. Additionally IF bandwidths of less than 1MHz
/// below 50MHz FCentre exhibit significant distortion and their use is not
/// recommended. This limitation will be addressed in planned future releases.
/// @param	*pStream		Stream handle (returned on success)
/// @param	fCentreHz		Centre frequency of frames to be returned by this stream
/// @param	fIFBWHz			IF Bandwidth of frames to be returned by this stream
/// @param	fRefLeveldBm	Reference Level in dBm of maximum expected signal
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet IQCaptureExample.cpp trfCreateIQStream
//-------------------------------------
_TRFAPI trfStatus trfCreateIQStream(
	trfHandle *pStream,
	_float32 fCentreHz, _float32 fIFBWHz,
	_float32 fRefLeveldBm
);

// trfGetNextIQ
/// Obtain the next IQ frame (if available) from the stream. If no frame is yet
/// available trfStatus will be 'trfWaiting'. If the stream has been halted
/// then the status 'trfExhausted' may be returned to indicate there are
/// no more buffered frames to retrieve. If the retrieval succeeds,
/// trfStatus will be trfOk and *ppFrame will point to the returned
/// IQ frame. The frame should be disposed of using trfFreeIQ.
/// The trfExhausted return indicates that capture for this 'attach' is complete
/// and the stream has been detached.
/// @param	cStream			Stream handle
/// @param	ppFrame			Address of pointer to frame to recieve IQ data.
/// @param  uTimeoutMsec	Milliseconds to wait for the next packet
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet IQCaptureExample.cpp trfGetNextIQ
//-------------------------------------
_TRFAPI trfStatus trfGetNextIQ(trfHandle cStream, trfIQFrame **ppFrame, _uint32 uTimeoutMsec);

// trfFreeIQ
/// Free the given IQ frame, which will no longer be valid
/// after this call (NULL). The user should ensure that frames
/// are disposed when no longer required to avoid 
/// STORAGE_CAPACITY_VIOLATION errors.
/// @param	ppFrame		Address of pointer to IQ frame to dispose of
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfFreeIQ(trfIQFrame **ppFrame);

// trfCreateSpectrumStream
/// Create a spectrum stream with the given parameters. This entity will be inactive
/// until successfully attached to a device capable of servicing the stream.
/// @param	*pStream		Stream handle (returned on success)
/// @param	fMinHz			Minimum extent of swept frequency range
/// @param	fMaxHz			Maximum extent of swept frequency range
/// @param	fRBWHz			Resolution bandwidth of returned sweeps
/// @param	pWindowType		Window type to use for spectrum reconstruction (nullptr - default)
/// @param	fRefLeveldBm	Reference Level in dBm of maximum expected signal
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet SpectrumCaptureExample.cpp trfCreateSpectrumStream
//-------------------------------------
_TRFAPI trfStatus trfCreateSpectrumStream(
	trfHandle *pStream,
	_float32 fMinHz, _float32 fMaxHz, _float32 fRBWHz,
	const char *pWindowType, _float32 fRefLeveldBm
);

// trfGetNextSpectrum
/// Obtain the next frame (if available) from the stream. If no frame is yet
/// available trfStatus will be 'WAITING' and *ppFrame will be NULL otherwise
/// trfStatus will be trfOk and *ppFrame will point to the returned
/// data frame. The frame should be disposed of using trfFreeSpectrum.
/// @param	cStream			Stream handle
/// @param	ppFrame			Address of pointer to receive frame pointer
/// @param	uTimeoutMsec	Maximum time in milliseconds to wait for frame
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet SpectrumCaptureExample.cpp trfGetNextSpectrum
//-------------------------------------
_TRFAPI trfStatus trfGetNextSpectrum(trfHandle cStream, trfSpectrumFrame **ppFrame, _uint32 uTimeoutMsec);

// trfFreeSpectrum
/// Free the given spectrum frame, which will no longer be valid
/// after this call. The user should ensure that frames are disposed
/// when no longer required to avoid STORAGE_CAPACITY_VIOLATION
/// errors.
/// @param	ppFrame		Address of pointer to Spectrum frame to dispose of
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfFreeSpectrum(trfSpectrumFrame **ppFrame);

// trfGetNextBaseband
/// Obtain the next Baseband frame (if available) from the stream. If no frame is yet
/// available trfStatus will be 'trfWaiting'. If the stream has been halted
/// then the status 'trfExhausted' may be returned to indicate there are
/// no more buffered frames to retrieve. If the retrieval succeeds,
/// trfStatus will be trfOk and *ppFrame will point to the returned
/// Baseband frame. The frame should be disposed of using trfFreeBaseband.
/// The trfExhausted return indicates that capture for this 'attach' is complete
/// and the stream has been detached.
/// @param	cStream			Stream handle
/// @param	ppFrame			Address of pointer to frame to recieve IQ data.
/// @param  uTimeoutMsec	Milliseconds to wait for the next packet
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfGetNextBaseband
//-------------------------------------
_TRFAPI trfStatus trfGetNextBaseband(trfHandle cStream, trfBasebandFrame **ppFrame, _uint32 uTimeoutMsec);

// trfFreeBaseband
/// Free the given IQ frame, which will no longer be valid
/// after this call (NULL). The user should ensure that frames
/// are disposed when no longer required to avoid 
/// STORAGE_CAPACITY_VIOLATION errors.
/// @param	ppFrame		Address of pointer to IQ frame to dispose of
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfFreeBaseband(trfBasebandFrame **ppFrame);


//*** Stream manipulation
//*****************************************************************************
// trfAttachStreamToDevice
/// Attach the given stream to the specified device. A stream may only be
/// attached to one device at a time. If the device supports the stream, the
/// call will return trfOk and the stream will start to produce data, subject
/// to resource availability on the specified device.
/// @param cDevice		Device handle
/// @param cStream		Stream handle
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet IQCaptureExample.cpp trfAttachStreamToDevice
//-------------------------------------
_TRFAPI trfStatus trfAttachStreamToDevice(trfHandle cDevice, trfHandle cStream);

// trfDetachStream
/// Detach the given stream from any device to which it is connected. This
/// will stop any capture associated with the stream - however the stream will
/// still contain any buffered data that has been previously delivered.
/// @param cStream		Receives stream handle on success
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet IQCaptureExample.cpp trfDetachStream
//-------------------------------------
_TRFAPI trfStatus trfDetachStream(trfHandle cStream);

// trfFlushStream
/// Remove all frames currently present in a stream. If the stream is attached
/// and delivering data, this will potentially create a break in the data. In
/// the case that the stream is detached, this will remove any outstanding
/// frames. After this call any call to trfGetNextXXX will return trfExhausted.
/// @param cStream		Receives stream handle on success
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfFlushStream(trfHandle cStream);

// trfIsAttached
/// Test if the specified stream is attached. Note that a finite capture
/// stream will detach itself after capture is completed. This change will
/// occur at some point in time after the completion of capture, and hence
/// it is valid for this function to return trfOk whilst an associated 
/// trfGetNextXXXFrame has returned trfExhausted (finite capture).
/// @param cStream		Stream handle
/// @return error/warning information. If attached returns trfOk. If detached
/// returns the warning trfDetached. trfBadHandle may also be returned.
/// <p> @b Example: @snippet FiniteIQCaptureExample.cpp trfIsAttached
//-------------------------------------
_TRFAPI trfStatus trfIsAttached(trfHandle cStream);

// trfAllowDiscarding
/// By default, the data contained in streams will be discarded to ensure
/// buffer size constraints are respected. In certain cases - such as streaming
/// from file - it is necessary to control the flow of data to keep a
/// processing chain synchronized and not arbitrarily drop frames. This call
/// allows streams that would normally contain non-discardable frames to allow
/// discarding to occur. By setting this flag on (say) an IQ stream from a file
/// that is being processed downstream, the user will not have to loop to keep
/// this stream emptied to allow the processor to proceed - otherwise the stream
/// would block waiting for the stream buffer to be emptied.
/// @param cStream				Stream handle
/// @param bAllowDiscarding		If set true, then this ensures normal discarding
///								behaviour is followed.
/// @return error/warning information. Expect trfOk.
/// <p> @b Example: @snippet IQToSpectrumExampleFromFile.cpp trfAllowDiscarding
//-------------------------------------
_TRFAPI trfStatus trfAllowDiscarding(trfHandle cStream, bool bAllowDiscarding);


//*** Stream file handling
//*****************************************************************************
// trfSetStreamOutputFile
/// Attaches the specified output file - or if the file exists and is not
/// empty, a numbered sibling file to the specified stream. The self-contained output file 
/// may be replayed using the trfCreateFileSourceStream function. 
/// The output file shall have the extension '.data' (ThinkRF Spectrum, IQ, Baseband),
/// The binary file shall each be accompanied by a metadata file (.json).
/// If pFilePath is NULL, any file writing for the stream shall be immediately halted.
/// @param	cStream		Stream handle
/// @param	pFilePath	Fully qualified output path for data file.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet CaptureToFileExample.cpp trfSetStreamOutputFile
//-------------------------------------
_TRFAPI trfStatus trfSetStreamOutputFile(trfHandle cStream, const char *pFilePath);

// trfAttachStreamToFile
/// Attaches the given stream to an input file. Assumes that the file is valid and
/// of the correct type for the stream - otherwise the call will fail. Playback
/// from the stream is controllable through the following parameters that may
/// be set for the stream:
/// TRFLooping - boolean: if set true then the file will be played continuously
///			   by looping back to the start once the file has been completely
///			   delivered.
/// Function parameters:
/// @param	cStream		Handle to stream to attach to file
/// @param	*pFilePath	Fully qualified input path for json/data file.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet CaptureToFileExample.cpp trfAttachStreamToFile
//-------------------------------------
_TRFAPI trfStatus trfAttachStreamToFile(trfHandle cStream, const char *pFilePath);

// trfConvertBasebandFileToWavFile
/// Stand-alone function to convert a previously created Baseband stream file and produce
/// a WAV file equivalent output file.
/// @param	*pSourceFilePath		Fully qualified path to previously generated Baseband data file.
/// @param	*pDestinationFilePath	Fully qualified output path for equivalent wav file.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfConvertBasebandFileToWavFile
//-------------------------------------
_TRFAPI trfStatus trfConvertBasebandFileToWavFile(const char *pSourceFilePath, const char *pDestinationFilePath);


//*** Processor Handling
//*****************************************************************************
// trfGetProcessorTypes
/// Returns a JSON-formatted array names 'processortypes' of the available
/// processor types.
/// @param	**ppJSON		*pJSON contains the list of available processors on exit.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfGetProcessorTypes
//-------------------------------------
_TRFAPI trfStatus trfGetProcessorTypes(char **ppJSON);

// trfCreateProcessor
/// Create a processor entity of the given type. Returns a handle to the newly
/// created processor.
/// @param	*pProcessor		Processor handle (returned on success)
/// @param	*pType			Textual type of processor to be created
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfCreateProcessor
//-------------------------------------
_TRFAPI trfStatus trfCreateProcessor(trfHandle *pProcessor, const char *pType);

// trfAttachProcessorToStream
/// Attach a processor to a source stream.
/// @param	hProcessor		Processor handle
/// @param	hSourceStream	Stream handle
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfAttachProcessorToStream
//-------------------------------------
_TRFAPI trfStatus trfAttachProcessorToStream(trfHandle hProcessor, trfHandle hSourceStream);

// trfDetachProcessor
/// Detach a processor from its current source stream.
/// @param	hProcessor		Processor handle
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfDetachProcessor
//-------------------------------------
_TRFAPI trfStatus trfDetachProcessor(trfHandle hProcessor);

// trfGetProcessorOutputStream
/// Return the output stream from the given processor.
/// @param	cProcessor	Processor handle
/// @param	pStream		Pointer to trfHandle to receive stream handle.
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfGetProcessorOutputStream
//-------------------------------------
_TRFAPI trfStatus trfGetProcessorOutputStream(trfHandle cProcessor, trfHandle *pStream);


//*** Entity parameter control
//*****************************************************************************
// trfGetParameterInfo
/// Obtains all available parameters for the specified device along with the
/// valid ranges, type and any associated information through ppJSON.
/// @param cEntity		Device/Processor/Stream handle
/// @param ppJSON		Pointer to receive UTF8 JSON string
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfGetParameterInfo
//-------------------------------------
_TRFAPI trfStatus trfGetParameterInfo(trfHandle cEntity, char **ppJSON);

// trfReadParameterInfoElement
/// Read a sub-element from a JSON element as a _float32 value. Useful for
/// obtaining 'min', 'max' and 'default' values from parameter information.
/// @param ppJSON		Source JSON string
/// @param pParameter	Name of parameter to be extracted from
/// @param pElement		Name of element (ie. "min", "max", "default")
/// @param pfValue		Pointer to _float32 to receive value
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadParameterInfoElement(
	char **ppJSON, const char *pParameter, const char *pElement, _float32 *pfValue
);

// trfReadParameterAsJSON
/// Read a sub-element from a JSON element as another JSON string. Useful for
/// extracting substructure from a JSON string.
/// @param pJSON			Source JSON string
/// @param pParameter		Name of parameter to extract
/// @param ppJSONExtract	Pointer to receive JSON string
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadParameterAsJSON(
	char **pJSON, const char *pParameter, char **ppJSONExtract
);

// trfGetParameters
/// Obtain the current values of device/processor parameters and make accessible through
/// ppJSON.
/// @param cEntity	Device/Processor/Stream handle
/// @param ppJSON	Pointer to receive UTF8 JSON string describing parameters
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet CaptureToFileExample.cpp trfGetParameters
//-------------------------------------
_TRFAPI trfStatus trfGetParameters(trfHandle cEntity, char **ppJSON);

// trfSetParameters
/// Set the current value of a device/stream/processor parameter.
/// @param cEntity	Device/Processor/Stream handle
/// @param ppJSON	Pointer to JSON string pointer, describing parameter(s) to be set.
///					On return *ppJSON will be nullptr and storage will be disposed of.
/// @return error/warning information (on success == trfOk)
/// <p> @b Example: @snippet AMDemodulationExample.cpp trfSetParameters
//-------------------------------------
_TRFAPI trfStatus trfSetParameters(trfHandle cEntity, char **ppJSON);

// trfDisposeString
/// Dispose of a previously allocated ppJSON string. Note that only strings that
/// are returned via a parameter of the form 'char **' should be disposed of using
/// this call. This is provided for syntactic convenience and is equivalent to:
/// if (ppJSON != nullptr) {
///   if (*ppJSON != nullptr) {
///	 	delete [](*ppJSON);
///		*ppJSON = nullptr;
///	  }
/// }
//-------------------------------------
_TRFAPI void trfDisposeString(char **ppString);

// trfAddNumericParameter
/// For parameters not named above - add parameter by name
/// Add named numeric parameter to ppJSON
/// @param ppJSON			Pointer to JSON string - accumulating parameters.
/// @param pParameterName	Name of parameter to add.
/// @param fValue			Numeric value to add.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddNumericParameter(char **ppJSON, const char *pParameterName, _float32 fValue);

// trfAddTextParameter
/// Add named textual parameter to ppJSON
/// @param ppJSON			Pointer to JSON string - accumulating parameters.
/// @param pParameterName	Name of parameter to add.
/// @param pValue			Textual value to add (zero-terminated string).
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddTextParameter(char **ppJSON, const char *pParameterName, const char *pValue);

// trfAddBooleanParameter
/// Add named boolean parameter to ppJSON
/// @param ppJSON			Pointer to JSON string - accumulating parameters.
/// @param pParameterName	Name of parameter to add.
/// @param bValue			Boolean value to add.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddBooleanParameter(char **ppJSON, const char *pParameterName, bool bValue);

// trfReadNumericParameter
/// Read named numeric parameter from ppJSON
/// @param ppJSON			Pointer to JSON string containing parameter(s).
/// @param pParameterName	Name of numeric parameter to read.
/// @param pfValue			Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadNumericParameter(char **ppJSON, const char *pParameterName, _float32 *pfValue);

// trfReadTextParameter
/// Read named textual parameter from ppJSON
/// @param ppJSON			Pointer to JSON string containing parameter(s).
/// @param pParameterName	Name of textual parameter to read.
/// @param ppValue			Pointer to char pointer to receive value - allocated as []char.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadTextParameter(char **ppJSON, const char *pParameterName, char **ppValue);

// trfReadBooleanParameter
/// Read named boolean parameter from ppJSON
/// @param ppJSON			Pointer to JSON string containing parameter(s).
/// @param pParameterName	Name of numeric parameter to read.
/// @param pbValue			Pointer to bool to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadBooleanParameter(char **ppJSON, const char *pParameterName, bool *pbValue);


// Syntactic sugar for common device parameters to aid in constructing
// ppJSON to be used with trfSetParameters.
//*************************************
// trfAddFCentre
/// Add Centre frequency parameter to ppJSON. Centre frequency parameter
/// is required by IQ Streams.
/// @param ppJSON		Pointer to JSON string - accumulating parameters.
/// @param fFCentreHz	Centre frequency to add to JSON parameter string.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddFCentre(char **ppJSON, _float32 fFCentreHz);

// trfAddIFBWHz
/// Add IF Bandwidth (IFBW) parameter to ppJSON. IFBW is required by IQ Streams.
/// @param ppJSON		Pointer to JSON string - accumulating parameters.
/// @param fFIFBWHz		IF Bandwidth parameter to add to JSON parameter string.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddIFBWHz(char **ppJSON, _float32 fFIFBWHz);	 

// trfAddFMinMax
/// Add min/max frequency range to ppJSON. The min and max frequencies specified
/// are required by Spectrum streams.
/// @param ppJSON		Pointer to JSON string - accumulating parameters.
/// @param fFMinHz		Minimum extent frequency to add to JSON parameter string.
/// @param fFMaxHz		Maximum extent frequency to add to JSON parameter string.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddFMinMax(char **ppJSON, _float32 fFMinHz, _float32 fFMaxHz);	

// trfAddRBWHz
/// Add resolution bandwidth parameter to ppJSON. RBW parameter is required by spectrum
/// streams.
/// @param ppJSON		Pointer to JSON string - accumulating parameters.
/// @param fRBWHz		Resolution Bandwidth parameter to add to JSON parameter string.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddRBWHz(char **ppJSON, _float32 fRBWHz);				

// trfAddWindowType
/// Add window type specification to ppJSON. A window type is optional for spectrum
/// streams.
/// @param ppJSON		Pointer to JSON string - accumulating parameters.
/// @param pWindow		Textual window type to add to JSON parameter string.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddWindowType(char **ppJSON, const char *pWindow);		

// trfAddRefLeveldBm
/// Add reference level to ppJSON
/// The reference level is the strongest expected signal. Setting the
/// reference level instructs a receiver to optimize its signal path,
/// attenuation, gain settings to maximize available dynamic range for
/// this expected signal level. A reference level is optional but recommended
/// for IQ and Spectrum streams.
/// @param ppJSON		Pointer to JSON string - accumulating parameters.
/// @param fRefLeveldBm	Level in dBm of strongest expected signal.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddRefLeveldBm(char **ppJSON, _float32 fRefLeveldBm);

// trfAddIQCaptureSec
/// Add capture duration parameter ppJSON
/// The duration is expressed in seconds and for an IQ stream represents
/// the amount of data in seconds to be captured. After this capture
/// period has expired from attaching to a device, the stream will
/// automatically detach from the device and report 'trfExhausted' once
/// all captured data has been retrieved.
/// All devices will attempt to meet fIQCaptureSec at best-effort. In
/// the case that the captured time will fit in internal buffering, the
/// captured IQ sequence is guaranteed to be contiguous. If a stream duration
/// is short enough to be fully buffered, the read-only stream parameter
/// 'contiguous' will be present (and reads as 'true' (boolean)). A stream
/// for which the 'contiguous' flag is not present may contain discontinuities.
/// The read-only parameter 'maxContiguousSec' (_float32) gives the maximum
/// capture length that can fit in device buffering and thus for which the
/// 'contiguous' flag will be set with the current capture parameters.
/// @param ppJSON			Pointer to JSON string - accumulating parameters.
/// @param fIQCaptureSec	Number of seconds of data to capture.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfAddIQCaptureSec(char **ppJSON, _float32 fIQCaptureSec);

// trfReadFCentre
/// Read Centre frequency parameter from ppJSON
/// @param ppJSON		Pointer to JSON string containing parameter(s).
/// @param pfFCentreHz	Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadFCentre(char **ppJSON, _float32 *pfFCentreHz);

// trfReadIFBWHz
/// Read IFBW parameter from ppJSON
/// @param ppJSON		Pointer to JSON string containing parameter(s).
/// @param pfFIFBWHz	Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadIFBWHz(char **ppJSON, _float32 *pfFIFBWHz);

// trfReadFMinMax
/// Read min/max frequency range from ppJSON
/// @param ppJSON		Pointer to JSON string containing parameter(s).
/// @param pfFMinHz		Pointer to _float32 to receive value.
/// @param pfFMaxHz		Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadFMinMax(char **ppJSON, _float32 *pfFMinHz, _float32 *pfFMaxHz);

// trfReadRBWHz
/// Read resolution bandwidth parameter from ppJSON
/// @param ppJSON		Pointer to JSON string containing parameter(s).
/// @param pfRBWHz		Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadRBWHz(char **ppJSON, _float32 *pfRBWHz);

// trfReadWindowType
/// Read window type parameter from ppJSON
/// @param ppJSON		Pointer to JSON string containing parameter(s).
/// @param ppWindow		Pointer to char pointer to receive value - allocated as []char.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadWindowType(char **ppJSON, char **ppWindow);

// trfReadSampleRateHz
/// Read sample rate parameter from ppJSON
/// @param ppJSON			Pointer to JSON string containing parameter(s).
/// @param pfSampleRateHz	Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadSampleRateHz(char **ppJSON, _float32 *pfSampleRateHz);

// trfReadRefLeveldBm
/// Read reference level parameter from ppJSON
/// @param ppJSON			Pointer to JSON string containing parameter(s).
/// @param pfRefLeveldBm	Pointer to _float32 to receive value.
/// @return error/warning information (on success == trfOk)
//-------------------------------------
_TRFAPI trfStatus trfReadRefLeveldBm(char **ppJSON, _float32 *pfRefLeveldBm);

#endif // _LIBTRF_
